from hummingbot.core.api_throttler.data_types import RateLimit
from hummingbot.core.data_type.common import OrderType, PositionMode
from hummingbot.core.data_type.in_flight_order import OrderState

EXCHANGE_NAME = "bitget_perpetual"

DEFAULT_DOMAIN = ""

DEFAULT_TIME_IN_FORCE = "normal"

REST_URL = "https://api.bitget.com"
WSS_URL = "wss://ws.bitget.com/mix/v1/stream"

SECONDS_TO_WAIT_TO_RECEIVE_MESSAGE = 20  # According to the documentation this has to be less than 30 seconds

ORDER_TYPE_MAP = {
    OrderType.LIMIT: "limit",
    OrderType.MARKET: "market",
}

POSITION_MODE_API_ONEWAY = "fixed"
POSITION_MODE_API_HEDGE = "crossed"
POSITION_MODE_MAP = {
    PositionMode.ONEWAY: POSITION_MODE_API_ONEWAY,
    PositionMode.HEDGE: POSITION_MODE_API_HEDGE,
}

SYMBOL_AND_PRODUCT_TYPE_SEPARATOR = "_"
USDT_PRODUCT_TYPE = "UMCBL"
USDC_PRODUCT_TYPE = "CMCBL"
USD_PRODUCT_TYPE = "DMCBL"
ALL_PRODUCT_TYPES = [USDT_PRODUCT_TYPE, USDC_PRODUCT_TYPE, USD_PRODUCT_TYPE]

# REST API Public Endpoints
LATEST_SYMBOL_INFORMATION_ENDPOINT = "/api/mix/v1/market/ticker"
QUERY_SYMBOL_ENDPOINT = "/api/mix/v1/market/contracts"
ORDER_BOOK_ENDPOINT = "/api/mix/v1/market/depth"
SERVER_TIME_PATH_URL = "/api/mix/v1/"
GET_LAST_FUNDING_RATE_PATH_URL = "/api/mix/v1/market/current-fundRate"
OPEN_INTEREST_PATH_URL = "/api/mix/v1/market/open-interest"
MARK_PRICE_PATH_URL = "/api/mix/v1/market/mark-price"

# REST API Private Endpoints
SET_LEVERAGE_PATH_URL = "/api/mix/v1/account/setLeverage"
GET_POSITIONS_PATH_URL = "/api/mix/v1/position/allPosition"
PLACE_ACTIVE_ORDER_PATH_URL = "/api/mix/v1/order/placeOrder"
CANCEL_ACTIVE_ORDER_PATH_URL = "/api/mix/v1/order/cancel-order"
CANCEL_ALL_ACTIVE_ORDERS_PATH_URL = "/api/mix/v1/order/cancel-batch-orders"
QUERY_ACTIVE_ORDER_PATH_URL = "/api/mix/v1/order/detail"
USER_TRADE_RECORDS_PATH_URL = "/api/mix/v1/order/fills"
GET_WALLET_BALANCE_PATH_URL = "/api/mix/v1/account/accounts"
SET_POSITION_MODE_URL = "/api/mix/v1/account/setMarginMode"
GET_FUNDING_FEES_PATH_URL = "/api/mix/v1/account/accountBill"

# Funding Settlement Time Span
FUNDING_SETTLEMENT_TIME_PATH_URL = "/api/mix/v1/market/funding-time"

# WebSocket Public Endpoints
WS_PING_REQUEST = "ping"
WS_PONG_RESPONSE = "pong"
WS_ORDER_BOOK_EVENTS_TOPIC = "books"
WS_TRADES_TOPIC = "trade"
WS_INSTRUMENTS_INFO_TOPIC = "tickers"
WS_AUTHENTICATE_USER_ENDPOINT_NAME = "login"
WS_SUBSCRIPTION_POSITIONS_ENDPOINT_NAME = "positions"
WS_SUBSCRIPTION_ORDERS_ENDPOINT_NAME = "orders"
WS_SUBSCRIPTION_WALLET_ENDPOINT_NAME = "account"

# Order Statuses
ORDER_STATE = {
    "new": OrderState.OPEN,
    "filled": OrderState.FILLED,
    "full-fill": OrderState.FILLED,
    "partial-fill": OrderState.PARTIALLY_FILLED,
    "partially_filled": OrderState.PARTIALLY_FILLED,
    "canceled": OrderState.CANCELED,
    "cancelled": OrderState.CANCELED,
}

# Request error codes
RET_CODE_OK = "00000"
RET_CODE_PARAMS_ERROR = "40007"
RET_CODE_API_KEY_INVALID = "40006"
RET_CODE_AUTH_TIMESTAMP_ERROR = "40005"
RET_CODE_ORDER_NOT_EXISTS = "43025"
RET_CODE_API_KEY_EXPIRED = "40014"


RATE_LIMITS = [
    RateLimit(
        limit_id=LATEST_SYMBOL_INFORMATION_ENDPOINT,
        limit=20,
        time_interval=1,
    ),
    RateLimit(
        limit_id=QUERY_SYMBOL_ENDPOINT,
        limit=20,
        time_interval=1,
    ),
    RateLimit(
        limit_id=ORDER_BOOK_ENDPOINT,
        limit=20,
        time_interval=1,
    ),
    RateLimit(
        limit_id=SERVER_TIME_PATH_URL,
        limit=20,
        time_interval=1,
    ),
    RateLimit(
        limit_id=GET_LAST_FUNDING_RATE_PATH_URL,
        limit=20,
        time_interval=1,
    ),
    RateLimit(
        limit_id=OPEN_INTEREST_PATH_URL,
        limit=20,
        time_interval=1,
    ),
    RateLimit(
        limit_id=MARK_PRICE_PATH_URL,
        limit=20,
        time_interval=1,
    ),
    RateLimit(
        limit_id=FUNDING_SETTLEMENT_TIME_PATH_URL,
        limit=20,
        time_interval=1,
    ),
    RateLimit(
        limit_id=SET_LEVERAGE_PATH_URL,
        limit=5,
        time_interval=2,
    ),
    RateLimit(
        limit_id=GET_POSITIONS_PATH_URL,
        limit=5,
        time_interval=2,
    ),
    RateLimit(
        limit_id=PLACE_ACTIVE_ORDER_PATH_URL,
        limit=10,
        time_interval=1,
    ),
    RateLimit(
        limit_id=CANCEL_ACTIVE_ORDER_PATH_URL,
        limit=10,
        time_interval=1,
    ),
    RateLimit(
        limit_id=CANCEL_ALL_ACTIVE_ORDERS_PATH_URL,
        limit=10,
        time_interval=1,
    ),
    RateLimit(
        limit_id=QUERY_ACTIVE_ORDER_PATH_URL,
        limit=20,
        time_interval=1,
    ),
    RateLimit(
        limit_id=USER_TRADE_RECORDS_PATH_URL,
        limit=20,
        time_interval=2,
    ),
    RateLimit(
        limit_id=GET_WALLET_BALANCE_PATH_URL,
        limit=20,
        time_interval=2,
    ),
    RateLimit(
        limit_id=SET_POSITION_MODE_URL,
        limit=5,
        time_interval=1,
    ),
    RateLimit(
        limit_id=GET_FUNDING_FEES_PATH_URL,
        limit=10,
        time_interval=1,
    ),
]



import logging
from typing import Any, Dict, List, Optional

import pandas as pd

from hummingbot.core.network_iterator import NetworkStatus
from hummingbot.core.utils.tracking_nonce import get_tracking_nonce
from hummingbot.core.web_assistant.connections.data_types import RESTMethod
from hummingbot.data_feed.candles_feed.candles_base import CandlesBase
from hummingbot.data_feed.candles_feed.kucoin_perpetual_candles import constants as CONSTANTS
from hummingbot.logger import HummingbotLogger


class KucoinPerpetualCandles(CandlesBase):
    _logger: Optional[HummingbotLogger] = None
    _last_ws_message_sent_timestamp = 0
    _ping_interval = 0

    @classmethod
    def logger(cls) -> HummingbotLogger:
        if cls._logger is None:
            cls._logger = logging.getLogger(__name__)
        return cls._logger

    def __init__(self, trading_pair: str, interval: str = "1m", max_records: int = 150):
        self.symbols_dict = {}
        self.hb_base_asset = trading_pair.split("-")[0]
        self.quote_asset = trading_pair.split("-")[1]
        self.kucoin_base_asset = self.get_kucoin_base_asset()
        super().__init__(trading_pair, interval, max_records)

    def get_kucoin_base_asset(self):
        for hb_asset, kucoin_value in CONSTANTS.HB_TO_KUCOIN_MAP.items():
            return kucoin_value if hb_asset == self.hb_base_asset else self.hb_base_asset

    @property
    def name(self):
        return f"kucoin_perpetual_{self._trading_pair}"

    @property
    def rest_url(self):
        return CONSTANTS.REST_URL

    @property
    def wss_url(self):
        return f"{self._ws_url}?token={self._ws_token}"

    @property
    def health_check_url(self):
        return self.rest_url + CONSTANTS.HEALTH_CHECK_ENDPOINT

    @property
    def candles_url(self):
        return self.rest_url + CONSTANTS.CANDLES_ENDPOINT

    @property
    def candles_endpoint(self):
        return CONSTANTS.CANDLES_ENDPOINT

    @property
    def candles_max_result_per_rest_request(self):
        return CONSTANTS.MAX_RESULTS_PER_CANDLESTICK_REST_REQUEST

    @property
    def symbols_url(self):
        return self.rest_url + CONSTANTS.SYMBOLS_ENDPOINT

    @property
    def public_ws_url(self):
        return self.rest_url + CONSTANTS.PUBLIC_WS_DATA_PATH_URL

    @property
    def rate_limits(self):
        return CONSTANTS.RATE_LIMITS

    @property
    def intervals(self):
        return CONSTANTS.INTERVALS

    @property
    def candles_df(self) -> pd.DataFrame:
        df = pd.DataFrame(self._candles, columns=self.columns, dtype=float)
        return df.sort_values(by="timestamp", ascending=True)

    @property
    def _ping_payload(self):
        return {
            "type": "ping",
        }

    async def check_network(self) -> NetworkStatus:
        rest_assistant = await self._api_factory.get_rest_assistant()
        await rest_assistant.execute_request(url=self.health_check_url,
                                             throttler_limit_id=CONSTANTS.HEALTH_CHECK_ENDPOINT)
        return NetworkStatus.CONNECTED

    def get_exchange_trading_pair(self, trading_pair):
        return f"{self.kucoin_base_asset}-{self.quote_asset}"

    @property
    def _is_last_candle_not_included_in_rest_request(self):
        return False

    @property
    def _is_first_candle_not_included_in_rest_request(self):
        return False

    def _get_rest_candles_params(self,
                                 start_time: Optional[int] = None,
                                 end_time: Optional[int] = None,
                                 limit: Optional[int] = None) -> dict:
        """
        For API documentation, please refer to:
        https://www.kucoin.com/docs/rest/futures-trading/market-data/get-klines
        """
        granularity = CONSTANTS.GRANULARITIES[self.interval]
        now = self._round_timestamp_to_interval_multiple(self._time())
        granularity_limits = {
            1: 24,  # 1 minute granularity, 24 hours
            5: 10 * 24,  # 5 minutes granularity, 10 days
            15: 30 * 24,  # 15 minutes granularity, 30 days
            30: 60 * 24,  # 30 minutes granularity, 60 days
            60: 120 * 24,  # 1 hour granularity, 120 days
            120: 240 * 24,  # 2 hours granularity, 240 days
            240: 480 * 24,  # 4 hours granularity, 480 days
            480: 720 * 24  # 6 hours granularity, 720 days
        }
        if granularity in granularity_limits:
            max_duration = granularity_limits[granularity] * 60  # convert days to minutes
            if (now - start_time) / 60 >= max_duration:
                raise ValueError(
                    f"{granularity}m granularity candles are only available for the last {granularity_limits[granularity] // 24} days.")

        params = {
            "symbol": self.symbols_dict[f"{self.kucoin_base_asset}-{self.quote_asset}"],
            "granularity": CONSTANTS.GRANULARITIES[self.interval],
            "to": end_time * 1000,
        }
        return params

    def _parse_rest_candles(self, data: dict, end_time: Optional[int] = None) -> List[List[float]]:
        return [[self.ensure_timestamp_in_seconds(row[0]), row[1], row[2], row[3], row[4], row[5], 0., 0., 0., 0.]
                for row in data['data']]

    def ws_subscription_payload(self):
        topic_candle = f"{self.symbols_dict[self._ex_trading_pair]}_{CONSTANTS.INTERVALS[self.interval]}"
        payload = {
            "id": str(get_tracking_nonce()),
            "type": "subscribe",
            "topic": f"/contractMarket/limitCandle:{topic_candle}",
            "privateChannel": False,
            "response": True,
        }
        return payload

    def _parse_websocket_message(self, data: dict):
        candles_row_dict: Dict[str, Any] = {}
        if data.get("data") is not None:
            if "candles" in data["data"]:
                candles = data["data"]["candles"]
                candles_row_dict["timestamp"] = self.ensure_timestamp_in_seconds(int(candles[0]))
                candles_row_dict["open"] = candles[1]
                candles_row_dict["close"] = candles[2]
                candles_row_dict["high"] = candles[3]
                candles_row_dict["low"] = candles[4]
                candles_row_dict["volume"] = candles[5]
                candles_row_dict["quote_asset_volume"] = 0.
                candles_row_dict["n_trades"] = 0.
                candles_row_dict["taker_buy_base_volume"] = 0.
                candles_row_dict["taker_buy_quote_volume"] = 0.
                return candles_row_dict

    async def initialize_exchange_data(self) -> Dict[str, Any]:
        await self._get_symbols_dict()
        await self._get_ws_token()

    async def _get_symbols_dict(self):
        try:
            rest_assistant = await self._api_factory.get_rest_assistant()
            response = await rest_assistant.execute_request(url=self.symbols_url,
                                                            throttler_limit_id=CONSTANTS.SYMBOLS_ENDPOINT)
            symbols = response["data"]
            symbols_dict = {}
            for symbol in symbols:
                symbols_dict[f"{symbol['baseCurrency']}-{symbol['quoteCurrency']}"] = symbol["symbol"]
            self.symbols_dict = symbols_dict
        except Exception:
            self.logger().error("Error fetching symbols from Kucoin.")
            raise

    async def _get_ws_token(self):
        try:
            rest_assistant = await self._api_factory.get_rest_assistant()
            connection_info = await rest_assistant.execute_request(
                url=self.public_ws_url,
                method=RESTMethod.POST,
                throttler_limit_id=CONSTANTS.PUBLIC_WS_DATA_PATH_URL,
            )

            self._ws_url = connection_info["data"]["instanceServers"][0]["endpoint"]
            self._ping_timeout = int(connection_info["data"]["instanceServers"][0]["pingTimeout"]) * 1e-3
            self._ws_token = connection_info["data"]["token"]
        except Exception:
            self.logger().error("Error fetching WS token from Kucoin.")
            raise



import asyncio
import threading
import time
from collections import OrderedDict, deque
from typing import TYPE_CHECKING, Dict, List

import pandas as pd

from hummingbot import check_dev_mode
from hummingbot.client.command.gateway_command import GatewayCommand
from hummingbot.client.config.config_helpers import (
    ClientConfigAdapter,
    get_strategy_config_map,
    missing_required_configs_legacy,
)
from hummingbot.client.config.security import Security
from hummingbot.client.settings import ethereum_wallet_required, required_exchanges
from hummingbot.connector.connector_base import ConnectorBase
from hummingbot.core.network_iterator import NetworkStatus
from hummingbot.core.utils.async_utils import safe_ensure_future
from hummingbot.logger.application_warning import ApplicationWarning
from hummingbot.user.user_balances import UserBalances

if TYPE_CHECKING:
    from hummingbot.client.hummingbot_application import HummingbotApplication  # noqa: F401


class StatusCommand:
    def _expire_old_application_warnings(self,  # type: HummingbotApplication
                                         ):
        now: float = time.time()
        expiry_threshold: float = now - self.APP_WARNING_EXPIRY_DURATION
        while len(self._app_warnings) > 0 and self._app_warnings[0].timestamp < expiry_threshold:
            self._app_warnings.popleft()

    def _format_application_warnings(self,  # type: HummingbotApplication
                                     ) -> str:
        lines: List[str] = []
        if len(self._app_warnings) < 1:
            return ""

        lines.append("\n  Warnings:")

        if len(self._app_warnings) < self.APP_WARNING_STATUS_LIMIT:
            for app_warning in reversed(self._app_warnings):
                lines.append(f"    * {pd.Timestamp(app_warning.timestamp, unit='s')} - "
                             f"({app_warning.logger_name}) - {app_warning.warning_msg}")
        else:
            module_based_warnings: OrderedDict = OrderedDict()
            for app_warning in reversed(self._app_warnings):
                logger_name: str = app_warning.logger_name
                if logger_name not in module_based_warnings:
                    module_based_warnings[logger_name] = deque([app_warning])
                else:
                    module_based_warnings[logger_name].append(app_warning)

            warning_lines: List[str] = []
            while len(warning_lines) < self.APP_WARNING_STATUS_LIMIT:
                logger_keys: List[str] = list(module_based_warnings.keys())
                for key in logger_keys:
                    warning_item: ApplicationWarning = module_based_warnings[key].popleft()
                    if len(module_based_warnings[key]) < 1:
                        del module_based_warnings[key]
                    warning_lines.append(f"    * {pd.Timestamp(warning_item.timestamp, unit='s')} - "
                                         f"({key}) - {warning_item.warning_msg}")
            lines.extend(warning_lines[:self.APP_WARNING_STATUS_LIMIT])

        return "\n".join(lines)

    async def strategy_status(self, live: bool = False):
        active_paper_exchanges = [exchange for exchange in self.markets.keys() if exchange.endswith("paper_trade")]

        paper_trade = "\n  Paper Trading Active: All orders are simulated, and no real orders are placed." if len(active_paper_exchanges) > 0 \
            else ""
        if asyncio.iscoroutinefunction(self.strategy.format_status):
            st_status = await self.strategy.format_status()
        else:
            st_status = self.strategy.format_status()
        status = paper_trade + "\n" + st_status
        return status

    def application_warning(self):
        # Application warnings.
        self._expire_old_application_warnings()
        if check_dev_mode() and len(self._app_warnings) > 0:
            app_warning = self._format_application_warnings()
            self.notify(app_warning)
            return app_warning

    async def validate_required_connections(
        self  # type: HummingbotApplication
    ) -> Dict[str, str]:
        invalid_conns = {}
        if not any([str(exchange).endswith("paper_trade") for exchange in required_exchanges]):
            if any([UserBalances.instance().is_gateway_market(exchange) for exchange in required_exchanges]):
                connections = await GatewayCommand.update_exchange(self, self.client_config_map, exchanges=required_exchanges)
            else:
                connections = await UserBalances.instance().update_exchanges(self.client_config_map, exchanges=required_exchanges)
            invalid_conns.update({ex: err_msg for ex, err_msg in connections.items()
                                  if ex in required_exchanges and err_msg is not None})
            if ethereum_wallet_required():
                err_msg = UserBalances.validate_ethereum_wallet()
                if err_msg is not None:
                    invalid_conns["ethereum"] = err_msg
        return invalid_conns

    def missing_configurations_legacy(
        self,  # type: HummingbotApplication
    ) -> List[str]:
        config_map = self.strategy_config_map
        missing_configs = []
        if not isinstance(config_map, ClientConfigAdapter):
            missing_configs = missing_required_configs_legacy(
                get_strategy_config_map(self.strategy_name)
            )
        return missing_configs

    def validate_configs(
        self,  # type: HummingbotApplication
    ) -> List[str]:
        config_map = self.strategy_config_map
        validation_errors = config_map.validate_model() if isinstance(config_map, ClientConfigAdapter) else []
        return validation_errors

    def status(self,  # type: HummingbotApplication
               live: bool = False):
        if threading.current_thread() != threading.main_thread():
            self.ev_loop.call_soon_threadsafe(self.status, live)
            return

        safe_ensure_future(self.status_check_all(live=live), loop=self.ev_loop)

    async def status_check_all(self,  # type: HummingbotApplication
                               notify_success=True,
                               live=False) -> bool:

        if self.strategy is not None:
            if live:
                await self.stop_live_update()
                self.app.live_updates = True
                while self.app.live_updates and self.strategy:
                    await self.cls_display_delay(
                        await self.strategy_status(live=True) + "\n\n Press escape key to stop update.", 0.1
                    )
                self.app.live_updates = False
                self.notify("Stopped live status display update.")
            else:
                self.notify(await self.strategy_status())
            return True

        # Preliminary checks.
        self.notify("\nPreliminary checks:")
        if self.strategy_name is None or self.strategy_file_name is None:
            self.notify('  - Strategy check: Please import or create a strategy.')
            return False

        if not Security.is_decryption_done():
            self.notify('  - Security check: Encrypted files are being processed. Please wait and try again later.')
            return False

        missing_configs = self.missing_configurations_legacy()
        if missing_configs:
            self.notify("  - Strategy check: Incomplete strategy configuration. The following values are missing.")
            for config in missing_configs:
                self.notify(f"    {config.key}")
        elif notify_success:
            self.notify('  - Strategy check: All required parameters confirmed.')
        validation_errors = self.validate_configs()
        if len(validation_errors) != 0:
            self.notify("  - Strategy check: Validation of the config maps failed. The following errors were flagged.")
            for error in validation_errors:
                self.notify(f"    {error}")
            return False

        network_timeout = float(self.client_config_map.commands_timeout.other_commands_timeout)
        try:
            invalid_conns = await asyncio.wait_for(self.validate_required_connections(), network_timeout)
        except asyncio.TimeoutError:
            self.notify("\nA network error prevented the connection check to complete. See logs for more details.")
            raise
        if invalid_conns:
            self.notify('  - Exchange check: Invalid connections:')
            for ex, err_msg in invalid_conns.items():
                self.notify(f"    {ex}: {err_msg}")
        elif notify_success:
            self.notify('  - Exchange check: All connections confirmed.')

        if invalid_conns or missing_configs or len(validation_errors) != 0:
            return False

        loading_markets: List[ConnectorBase] = []
        for market in self.markets.values():
            if not market.ready:
                loading_markets.append(market)

        if len(loading_markets) > 0:
            self.notify("  - Connectors check:  Waiting for connectors "
                        f"{','.join([m.name.capitalize() for m in loading_markets])}"
                        " to get ready for trading. \n"
                        "                    Please keep the bot running and try to start again in a few minutes. \n")

            for market in loading_markets:
                market_status_df = pd.DataFrame(data=market.status_dict.items(), columns=["description", "status"])
                self.notify(
                    f"  - {market.display_name.capitalize()} connector status:\n" +
                    "\n".join(["     " + line for line in market_status_df.to_string(index=False,).split("\n")]) +
                    "\n"
                )
            return False

        elif not all([market.network_status is NetworkStatus.CONNECTED for market in self.markets.values()]):
            offline_markets: List[str] = [
                market_name
                for market_name, market
                in self.markets.items()
                if market.network_status is not NetworkStatus.CONNECTED
            ]
            for offline_market in offline_markets:
                self.notify(f"  - Connector check: {offline_market} is currently offline.")
            return False

        self.notify("  - All checks: Confirmed.")
        return True



import asyncio
import time
from typing import TYPE_CHECKING, Any, Dict, List, Optional

from hummingbot.connector.exchange.kraken import kraken_constants as CONSTANTS, kraken_web_utils as web_utils
from hummingbot.connector.exchange.kraken.kraken_order_book import KrakenOrderBook
from hummingbot.connector.exchange.kraken.kraken_utils import (
    convert_from_exchange_trading_pair,
    convert_to_exchange_trading_pair,
)
from hummingbot.core.data_type.order_book import OrderBook
from hummingbot.core.data_type.order_book_message import OrderBookMessage
from hummingbot.core.data_type.order_book_tracker_data_source import OrderBookTrackerDataSource
from hummingbot.core.web_assistant.connections.data_types import RESTMethod, WSJSONRequest
from hummingbot.core.web_assistant.rest_assistant import RESTAssistant
from hummingbot.core.web_assistant.web_assistants_factory import WebAssistantsFactory
from hummingbot.core.web_assistant.ws_assistant import WSAssistant
from hummingbot.logger import HummingbotLogger

if TYPE_CHECKING:
    from hummingbot.connector.exchange.kraken.kraken_exchange import KrakenExchange


class KrakenAPIOrderBookDataSource(OrderBookTrackerDataSource):
    MESSAGE_TIMEOUT = 30.0

    # PING_TIMEOUT = 10.0

    def __init__(self,
                 trading_pairs: List[str],
                 connector: 'KrakenExchange',
                 api_factory: WebAssistantsFactory,
                 # throttler: Optional[AsyncThrottler] = None
                 ):
        super().__init__(trading_pairs)
        self._connector = connector
        self._api_factory = api_factory
        self._rest_assistant = None
        self._ws_assistant = None
        self._order_book_create_function = lambda: OrderBook()

    _kraobds_logger: Optional[HummingbotLogger] = None

    async def _get_rest_assistant(self) -> RESTAssistant:
        if self._rest_assistant is None:
            self._rest_assistant = await self._api_factory.get_rest_assistant()
        return self._rest_assistant

    async def get_last_traded_prices(self,
                                     trading_pairs: List[str],
                                     domain: Optional[str] = None) -> Dict[str, float]:
        return await self._connector.get_last_traded_prices(trading_pairs=trading_pairs)

    async def _order_book_snapshot(self, trading_pair: str) -> OrderBook:
        snapshot: Dict[str, Any] = await self._request_order_book_snapshot(trading_pair)
        snapshot_timestamp: float = time.time()
        snapshot_msg: OrderBookMessage = KrakenOrderBook.snapshot_message_from_exchange(
            snapshot,
            snapshot_timestamp,
            metadata={"trading_pair": trading_pair}
        )
        return snapshot_msg

    async def _request_order_book_snapshot(self, trading_pair: str, ) -> Dict[str, Any]:
        """
        Retrieves a copy of the full order book from the exchange, for a particular trading pair.

        :param trading_pair: the trading pair for which the order book will be retrieved

        :return: the response from the exchange (JSON dictionary)
        """
        params = {
            "pair": await self._connector.exchange_symbol_associated_to_pair(trading_pair=trading_pair)
        }

        rest_assistant = await self._api_factory.get_rest_assistant()
        response_json = await rest_assistant.execute_request(
            url=web_utils.public_rest_url(path_url=CONSTANTS.SNAPSHOT_PATH_URL),
            params=params,
            method=RESTMethod.GET,
            throttler_limit_id=CONSTANTS.SNAPSHOT_PATH_URL,
        )
        if len(response_json["error"]) > 0:
            raise IOError(f"Error fetching Kraken market snapshot for {trading_pair}. "
                          f"Error is {response_json['error']}.")
        data: Dict[str, Any] = next(iter(response_json["result"].values()))
        data = {"trading_pair": trading_pair, **data}
        data["latest_update"] = max([*map(lambda x: x[2], data["bids"] + data["asks"])], default=0.)
        return data

    async def _subscribe_channels(self, ws: WSAssistant):
        """
        Subscribes to the trade events and diff orders events through the provided websocket connection.

        :param ws: the websocket assistant used to connect to the exchange
        """
        try:
            trading_pairs: List[str] = []
            for tp in self._trading_pairs:
                # trading_pairs.append(convert_to_exchange_trading_pair(tp, '/'))
                symbol = convert_to_exchange_trading_pair(tp, '/')
                trading_pairs.append(symbol)
            trades_payload = {
                "event": "subscribe",
                "pair": trading_pairs,
                "subscription": {"name": 'trade'},
            }
            subscribe_trade_request: WSJSONRequest = WSJSONRequest(payload=trades_payload)

            order_book_payload = {
                "event": "subscribe",
                "pair": trading_pairs,
                "subscription": {"name": 'book', "depth": 1000},
            }
            subscribe_orderbook_request: WSJSONRequest = WSJSONRequest(payload=order_book_payload)

            await ws.send(subscribe_trade_request)
            await ws.send(subscribe_orderbook_request)

            self.logger().info("Subscribed to public order book and trade channels...")
        except asyncio.CancelledError:
            raise
        except Exception:
            self.logger().error("Unexpected error occurred subscribing to order book data streams.")
            raise

    def _channel_originating_message(self, event_message) -> str:
        channel = ""
        if type(event_message) is list:
            channel = self._trade_messages_queue_key if event_message[-2] == CONSTANTS.TRADE_EVENT_TYPE \
                else self._diff_messages_queue_key
        else:
            if event_message.get("errorMessage") is not None:
                err_msg = event_message.get("errorMessage")
                raise IOError(f"Error event received from the server ({err_msg})")
        return channel

    async def _connected_websocket_assistant(self) -> WSAssistant:
        ws: WSAssistant = await self._api_factory.get_ws_assistant()
        await ws.connect(ws_url=CONSTANTS.WS_URL,
                         ping_timeout=CONSTANTS.PING_TIMEOUT)
        return ws

    async def _parse_trade_message(self, raw_message: Dict[str, Any], message_queue: asyncio.Queue):

        trades = [
            {"pair": convert_from_exchange_trading_pair(raw_message[-1]), "trade": trade}
            for trade in raw_message[1]
        ]
        for trade in trades:
            trade_msg: OrderBookMessage = KrakenOrderBook.trade_message_from_exchange(trade)
            message_queue.put_nowait(trade_msg)

    async def _parse_order_book_diff_message(self, raw_message: Dict[str, Any], message_queue: asyncio.Queue):
        msg_dict = {"trading_pair": convert_from_exchange_trading_pair(raw_message[-1]),
                    "asks": raw_message[1].get("a", []) or raw_message[1].get("as", []) or [],
                    "bids": raw_message[1].get("b", []) or raw_message[1].get("bs", []) or []}
        msg_dict["update_id"] = max(
            [*map(lambda x: float(x[2]), msg_dict["bids"] + msg_dict["asks"])], default=0.
        )
        if "as" in raw_message[1] and "bs" in raw_message[1]:
            order_book_message: OrderBookMessage = (
                KrakenOrderBook.snapshot_ws_message_from_exchange(msg_dict, time.time())
            )
        else:
            order_book_message: OrderBookMessage = KrakenOrderBook.diff_message_from_exchange(
                msg_dict, time.time())
        message_queue.put_nowait(order_book_message)



from hummingbot.core.data_type.common import OrderType, PositionMode
from hummingbot.core.data_type.in_flight_order import OrderState

EXCHANGE_NAME = "bybit_perpetual"

DEFAULT_DOMAIN = "bybit_perpetual_main"

DEFAULT_TIME_IN_FORCE = "GTC"

REST_URLS = {
    "bybit_perpetual_main": "https://api.bybit.com/",
    "bybit_perpetual_testnet": "https://api-testnet.bybit.com/"
}
WSS_NON_LINEAR_PUBLIC_URLS = {
    "bybit_perpetual_main": "wss://stream.bybit.com/v5/public/inverse",
    "bybit_perpetual_testnet": "wss://stream-testnet.bybit.com/v5/public/inverse"}
WSS_NON_LINEAR_PRIVATE_URLS = WSS_NON_LINEAR_PUBLIC_URLS
WSS_LINEAR_PUBLIC_URLS = {
    "bybit_perpetual_main": "wss://stream.bybit.com/v5/public/linear",
    "bybit_perpetual_testnet": "wss://stream-testnet.bybit.com/v5/public/linear"}
WSS_LINEAR_PRIVATE_URLS = {
    "bybit_perpetual_main": "wss://stream.bybit.com/v5/private",
    "bybit_perpetual_testnet": "wss://stream-testnet.bybit.com/v5/private"
}
WS_HEARTBEAT_TIME_INTERVAL = 20.0

# unit in millisecond and default value is 5,000, to specify how long an HTTP request is valid.
# It is also used to prevent replay attacks.
# https://bybit-exchange.github.io/docs/v5/guide#parameters-for-authenticated-endpoints
X_API_RECV_WINDOW = str(50000)

X_API_SIGN_TYPE = str(2)


HBOT_BROKER_ID = "Hummingbot"

MAX_ID_LEN = 36
SECONDS_TO_WAIT_TO_RECEIVE_MESSAGE = 30
POSITION_IDX_ONEWAY = 0
POSITION_IDX_HEDGE_BUY = 1
POSITION_IDX_HEDGE_SELL = 2

ORDER_TYPE_MAP = {
    OrderType.LIMIT: "Limit",
    OrderType.MARKET: "Market",
}

POSITION_MODE_API_ONEWAY = 0
POSITION_MODE_API_HEDGE = 3
POSITION_MODE_MAP = {
    PositionMode.ONEWAY: POSITION_MODE_API_ONEWAY,
    PositionMode.HEDGE: POSITION_MODE_API_HEDGE,
}

# REST API Public Endpoints
LINEAR_MARKET = "linear"
NON_LINEAR_MARKET = "non_linear"

# Covers: Spot / USDT perpetual / USDC contract / Inverse contract / Option
LATEST_SYMBOL_INFORMATION_ENDPOINT = {
    LINEAR_MARKET: "v5/market/tickers",
    NON_LINEAR_MARKET: "v5/market/tickers"}

QUERY_SYMBOL_ENDPOINT = {
    LINEAR_MARKET: "v5/market/instruments-info",
    NON_LINEAR_MARKET: "v5/market/instruments-info"}
ORDER_BOOK_ENDPOINT = {
    LINEAR_MARKET: "v5/market/orderbook",
    NON_LINEAR_MARKET: "v5/market/orderbook"}
SERVER_TIME_PATH_URL = {
    LINEAR_MARKET: "v5/market/time",
    NON_LINEAR_MARKET: "v5/market/time"
}

# REST API Private Endpoints
SET_LEVERAGE_PATH_URL = {
    LINEAR_MARKET: "v5/position/set-leverage",
    NON_LINEAR_MARKET: "v5/position/set-leverage"}
GET_LAST_FUNDING_RATE_PATH_URL = {
    LINEAR_MARKET: "v5/account/transaction-log",
    NON_LINEAR_MARKET: "v5/account/contract-transaction-log"}
GET_POSITIONS_PATH_URL = {
    LINEAR_MARKET: "v5/position/list",
    NON_LINEAR_MARKET: "v5/position/list"}
PLACE_ACTIVE_ORDER_PATH_URL = {
    LINEAR_MARKET: "v5/order/create",
    NON_LINEAR_MARKET: "v5/order/create"}
CANCEL_ACTIVE_ORDER_PATH_URL = {
    LINEAR_MARKET: "v5/order/cancel",
    NON_LINEAR_MARKET: "v5/order/cancel"}
QUERY_ACTIVE_ORDER_PATH_URL = {
    LINEAR_MARKET: "v5/order/realtime",
    NON_LINEAR_MARKET: "v5/order/realtime"}
USER_TRADE_RECORDS_PATH_URL = {
    LINEAR_MARKET: "v5/execution/list",
    NON_LINEAR_MARKET: "v5/execution/list"}
GET_WALLET_BALANCE_PATH_URL = {
    LINEAR_MARKET: "v5/account/wallet-balance",
    NON_LINEAR_MARKET: "v5/account/wallet-balance"}
SET_POSITION_MODE_URL = {
    LINEAR_MARKET: "v5/position/switch-mode"}
GET_TRANSFERABLE_AMOUNT_PATH_URL = {
    LINEAR_MARKET: "v5/account/withdrawal",
}
# Funding Settlement Time Span
FUNDING_SETTLEMENT_DURATION = (5, 5)  # seconds before snapshot, seconds after snapshot

# WebSocket Public Endpoints
WS_PING_REQUEST = "ping"
WS_TRADES_TOPIC = "publicTrade"
WS_ORDER_BOOK_EVENTS_TOPIC = "orderbook.200"
WS_INSTRUMENTS_INFO_TOPIC = "tickers"

# WebSocket Private Endpoints
WS_AUTHENTICATE_USER_ENDPOINT_NAME = "auth"
WS_SUBSCRIPTION_POSITIONS_ENDPOINT_NAME = "position"
WS_SUBSCRIPTION_ORDERS_ENDPOINT_NAME = "order"
WS_SUBSCRIPTION_EXECUTIONS_ENDPOINT_NAME = "execution"
WS_SUBSCRIPTION_WALLET_ENDPOINT_NAME = "wallet"

# Order Statuses
ORDER_STATE = {
    "Created": OrderState.OPEN,
    "New": OrderState.OPEN,
    "Filled": OrderState.FILLED,
    "PartiallyFilled": OrderState.PARTIALLY_FILLED,
    "Cancelled": OrderState.CANCELED,
    "PendingCancel": OrderState.PENDING_CANCEL,
    "Rejected": OrderState.FAILED,
}

GET_LIMIT_ID = "GETLimit"
POST_LIMIT_ID = "POSTLimit"
GET_RATE = 49  # per second
POST_RATE = 19  # per second

NON_LINEAR_PRIVATE_BUCKET_100_LIMIT_ID = "NonLinearPrivateBucket100"
NON_LINEAR_PRIVATE_BUCKET_600_LIMIT_ID = "NonLinearPrivateBucket600"
NON_LINEAR_PRIVATE_BUCKET_75_LIMIT_ID = "NonLinearPrivateBucket75"
NON_LINEAR_PRIVATE_BUCKET_120_B_LIMIT_ID = "NonLinearPrivateBucket120B"
NON_LINEAR_PRIVATE_BUCKET_120_C_LIMIT_ID = "NonLinearPrivateBucket120C"

LINEAR_PRIVATE_BUCKET_100_LIMIT_ID = "LinearPrivateBucket100"
LINEAR_PRIVATE_BUCKET_600_LIMIT_ID = "LinearPrivateBucket600"
LINEAR_PRIVATE_BUCKET_75_LIMIT_ID = "LinearPrivateBucket75"
LINEAR_PRIVATE_BUCKET_120_A_LIMIT_ID = "LinearPrivateBucket120A"

# Request error codes
RET_CODE_OK = 0

RET_CODE_MODE_POSITION_NOT_EMPTY = 110024
RET_CODE_MODE_NOT_MODIFIED = 110025
RET_CODE_MODE_ORDER_NOT_EMPTY = 110028
RET_CODE_HEDGE_NOT_SUPPORTED = 110029

RET_CODE_LEVERAGE_NOT_MODIFIED = 110043

RET_CODE_ORDER_NOT_EXISTS = 110001
RET_MSG_ORDER_NOT_EXISTS = "Order does not exist"
RET_CODE_ORDER_NOT_FOUND = 175007
RET_MSG_ORDER_NOT_FOUND = "Order not found."

RET_CODE_PARAMS_ERROR = 10001
RET_CODE_API_KEY_INVALID = 10003
RET_CODE_INVALID_TIME = 10002
RET_CODE_API_KEY_EXPIRED = 33004
RET_CODE_POSITION_ZERO = 130125



import asyncio
import sys
from decimal import Decimal
from typing import TYPE_CHECKING, Any, Dict, List, Optional

from hummingbot.connector.derivative.bitget_perpetual import (
    bitget_perpetual_constants as CONSTANTS,
    bitget_perpetual_web_utils as web_utils,
)
from hummingbot.core.data_type.common import TradeType
from hummingbot.core.data_type.funding_info import FundingInfo, FundingInfoUpdate
from hummingbot.core.data_type.order_book_message import OrderBookMessage, OrderBookMessageType
from hummingbot.core.data_type.perpetual_api_order_book_data_source import PerpetualAPIOrderBookDataSource
from hummingbot.core.utils.async_utils import safe_gather
from hummingbot.core.web_assistant.connections.data_types import RESTMethod, WSJSONRequest, WSPlainTextRequest
from hummingbot.core.web_assistant.web_assistants_factory import WebAssistantsFactory
from hummingbot.core.web_assistant.ws_assistant import WSAssistant

if TYPE_CHECKING:
    from hummingbot.connector.derivative.bitget_perpetual.bitget_perpetual_derivative import BitgetPerpetualDerivative


class BitgetPerpetualAPIOrderBookDataSource(PerpetualAPIOrderBookDataSource):

    FULL_ORDER_BOOK_RESET_DELTA_SECONDS = sys.maxsize

    def __init__(
        self,
        trading_pairs: List[str],
        connector: 'BitgetPerpetualDerivative',
        api_factory: WebAssistantsFactory,
        domain: str = ""
    ):
        super().__init__(trading_pairs)
        self._connector = connector
        self._api_factory = api_factory
        self._domain = domain
        self._diff_messages_queue_key = "books"
        self._trade_messages_queue_key = "trade"
        self._funding_info_messages_queue_key = "ticker"
        self._pong_response_event = None

    async def get_last_traded_prices(self, trading_pairs: List[str], domain: Optional[str] = None) -> Dict[str, float]:
        return await self._connector.get_last_traded_prices(trading_pairs=trading_pairs)

    async def get_funding_info(self, trading_pair: str) -> FundingInfo:
        funding_info_response = await self._request_complete_funding_info(trading_pair)
        funding_info = FundingInfo(
            trading_pair=trading_pair,
            index_price=Decimal(funding_info_response["amount"]),
            mark_price=Decimal(funding_info_response["markPrice"]),
            next_funding_utc_timestamp=int(int(funding_info_response["fundingTime"]) * 1e-3),
            rate=Decimal(funding_info_response["fundingRate"]),
        )
        return funding_info

    async def _process_websocket_messages(self, websocket_assistant: WSAssistant):
        while True:
            try:
                await asyncio.wait_for(
                    super()._process_websocket_messages(websocket_assistant=websocket_assistant),
                    timeout=CONSTANTS.SECONDS_TO_WAIT_TO_RECEIVE_MESSAGE)
            except asyncio.TimeoutError:
                if self._pong_response_event and not self._pong_response_event.is_set():
                    # The PONG response for the previous PING request was never received
                    raise IOError("The user stream channel is unresponsive (pong response not received)")
                self._pong_response_event = asyncio.Event()
                await self._send_ping(websocket_assistant=websocket_assistant)

    def _channel_originating_message(self, event_message: Dict[str, Any]) -> str:
        channel = ""
        if event_message == CONSTANTS.WS_PONG_RESPONSE and self._pong_response_event:
            self._pong_response_event.set()
        elif "event" in event_message:
            if event_message["event"] == "error":
                raise IOError(f"Public channel subscription failed ({event_message})")
        elif "arg" in event_message:
            channel = event_message["arg"].get("channel")
            if channel == CONSTANTS.WS_ORDER_BOOK_EVENTS_TOPIC and event_message.get("action") == "snapshot":
                channel = self._snapshot_messages_queue_key

        return channel

    async def _parse_order_book_diff_message(self, raw_message: Dict[str, Any], message_queue: asyncio.Queue):
        data = raw_message.get("data", {})
        inst_id = raw_message["arg"]["instId"]
        trading_pair = await self._connector.trading_pair_associated_to_exchange_instrument_id(instrument_id=inst_id)

        for book in data:
            update_id = int(book["ts"])
            timestamp = update_id * 1e-3

            order_book_message_content = {
                "trading_pair": trading_pair,
                "update_id": update_id,
                "bids": book["bids"],
                "asks": book["asks"],
            }
            diff_message = OrderBookMessage(
                message_type=OrderBookMessageType.DIFF,
                content=order_book_message_content,
                timestamp=timestamp
            )

            message_queue.put_nowait(diff_message)

    async def _parse_order_book_snapshot_message(self, raw_message: Dict[str, Any], message_queue: asyncio.Queue):
        data = raw_message.get("data", {})
        inst_id = raw_message["arg"]["instId"]
        trading_pair = await self._connector.trading_pair_associated_to_exchange_instrument_id(instrument_id=inst_id)

        for book in data:
            update_id = int(book["ts"])
            timestamp = update_id * 1e-3

            order_book_message_content = {
                "trading_pair": trading_pair,
                "update_id": update_id,
                "bids": book["bids"],
                "asks": book["asks"],
            }
            snapshot_msg: OrderBookMessage = OrderBookMessage(
                message_type=OrderBookMessageType.SNAPSHOT,
                content=order_book_message_content,
                timestamp=timestamp
            )
            message_queue.put_nowait(snapshot_msg)

    async def _parse_trade_message(self, raw_message: Dict[str, Any], message_queue: asyncio.Queue):
        data = raw_message.get("data", [])
        inst_id = raw_message["arg"]["instId"]
        trading_pair = await self._connector.trading_pair_associated_to_exchange_instrument_id(instrument_id=inst_id)

        for trade_data in data:
            ts_ms = int(trade_data[0])
            trade_type = float(TradeType.BUY.value) if trade_data[3] == "buy" else float(TradeType.SELL.value)
            message_content = {
                "trade_id": ts_ms,
                "trading_pair": trading_pair,
                "trade_type": trade_type,
                "amount": trade_data[2],
                "price": trade_data[1],
            }
            trade_message = OrderBookMessage(
                message_type=OrderBookMessageType.TRADE,
                content=message_content,
                timestamp=ts_ms * 1e-3,
            )
            message_queue.put_nowait(trade_message)

    async def _parse_funding_info_message(self, raw_message: Dict[str, Any], message_queue: asyncio.Queue):
        entries = raw_message.get("data", [])
        inst_id = raw_message["arg"]["instId"]
        trading_pair = await self._connector.trading_pair_associated_to_exchange_instrument_id(instrument_id=inst_id)

        for entry in entries:
            info_update = FundingInfoUpdate(trading_pair)
            info_update.index_price = Decimal(entry["indexPrice"])
            info_update.mark_price = Decimal(entry["markPrice"])
            info_update.next_funding_utc_timestamp = int(entry["nextSettleTime"]) * 1e-3
            info_update.rate = Decimal(entry["capitalRate"])
            message_queue.put_nowait(info_update)

    async def _request_complete_funding_info(self, trading_pair: str) -> Dict[str, Any]:
        params = {
            "symbol": await self._connector.exchange_symbol_associated_to_pair(trading_pair),
        }

        rest_assistant = await self._api_factory.get_rest_assistant()
        endpoints = [
            CONSTANTS.GET_LAST_FUNDING_RATE_PATH_URL,
            CONSTANTS.OPEN_INTEREST_PATH_URL,
            CONSTANTS.MARK_PRICE_PATH_URL,
            CONSTANTS.FUNDING_SETTLEMENT_TIME_PATH_URL
        ]
        tasks = []
        for endpoint in endpoints:
            tasks.append(rest_assistant.execute_request(
                url=web_utils.get_rest_url_for_endpoint(endpoint=endpoint),
                throttler_limit_id=endpoint,
                params=params,
                method=RESTMethod.GET,
            ))
        results = await safe_gather(*tasks)
        funding_info = {}
        for result in results:
            funding_info.update(result["data"])
        return funding_info

    async def _connected_websocket_assistant(self) -> WSAssistant:
        ws: WSAssistant = await self._api_factory.get_ws_assistant()
        await ws.connect(
            ws_url=CONSTANTS.WSS_URL, message_timeout=CONSTANTS.SECONDS_TO_WAIT_TO_RECEIVE_MESSAGE
        )
        return ws

    async def _subscribe_channels(self, ws: WSAssistant):
        try:
            payloads = []

            for trading_pair in self._trading_pairs:
                symbol = await self._connector.exchange_symbol_associated_to_pair_without_product_type(
                    trading_pair=trading_pair
                )

                for channel in [
                    self._diff_messages_queue_key,
                    self._trade_messages_queue_key,
                    self._funding_info_messages_queue_key,
                ]:
                    payloads.append({
                        "instType": "mc",
                        "channel": channel,
                        "instId": symbol
                    })
            final_payload = {
                "op": "subscribe",
                "args": payloads,
            }
            subscribe_request = WSJSONRequest(payload=final_payload)
            await ws.send(subscribe_request)
            self.logger().info("Subscribed to public order book, trade and funding info channels...")
        except asyncio.CancelledError:
            raise
        except Exception:
            self.logger().exception("Unexpected error occurred subscribing to order book trading and delta streams...")
            raise

    async def _request_order_book_snapshot(self, trading_pair: str) -> Dict[str, Any]:
        """
        Retrieves a copy of the full order book from the exchange, for a particular trading pair.

        :param trading_pair: the trading pair for which the order book will be retrieved

        :return: the response from the exchange (JSON dictionary)
        """
        symbol = await self._connector.exchange_symbol_associated_to_pair(trading_pair=trading_pair)
        params = {
            "symbol": symbol,
            "limit": "100",
        }

        rest_assistant = await self._api_factory.get_rest_assistant()
        data = await rest_assistant.execute_request(
            url=web_utils.public_rest_url(path_url=CONSTANTS.ORDER_BOOK_ENDPOINT),
            params=params,
            method=RESTMethod.GET,
            throttler_limit_id=CONSTANTS.ORDER_BOOK_ENDPOINT,
        )

        return data

    async def _order_book_snapshot(self, trading_pair: str) -> OrderBookMessage:
        snapshot_response: Dict[str, Any] = await self._request_order_book_snapshot(trading_pair)
        snapshot_data: Dict[str, Any] = snapshot_response["data"]
        update_id: int = int(snapshot_data["timestamp"])
        snapshot_timestamp: float = update_id * 1e-3

        order_book_message_content = {
            "trading_pair": trading_pair,
            "update_id": update_id,
            "bids": [(price, amount) for price, amount in snapshot_data.get("bids", [])],
            "asks": [(price, amount) for price, amount in snapshot_data.get("asks", [])],
        }
        snapshot_msg: OrderBookMessage = OrderBookMessage(
            OrderBookMessageType.SNAPSHOT,
            order_book_message_content,
            snapshot_timestamp)

        return snapshot_msg

    async def _send_ping(self, websocket_assistant: WSAssistant):
        ping_request = WSPlainTextRequest(payload=CONSTANTS.WS_PING_REQUEST)
        await websocket_assistant.send(ping_request)



import logging
from abc import ABC, abstractmethod
from contextlib import contextmanager
from typing import Generator, Optional

from sqlalchemy.exc import SQLAlchemyError
from sqlalchemy.orm import Session

log = logging.getLogger(__name__)


class TransactionBase(ABC):
    @abstractmethod
    def get_new_session(self) -> Session:
        raise NotImplementedError

    @contextmanager
    def begin(self) -> Generator[Session, None, None]:
        sql_session: Optional[Session] = None
        try:
            sql_session = self.get_new_session()
            yield sql_session
            sql_session.commit()
        except SQLAlchemyError:
            log.error("Rolling back SQL session due to error.")
            try:
                if sql_session is not None:
                    sql_session.rollback()
            except Exception:
                log.error("Failed to rollback a transaction!", exc_info=True)
            # Propagate to caller, as it must be aware of it too.
            raise
        finally:
            if sql_session is not None:
                sql_session.close()



import asyncio
from decimal import Decimal
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Tuple, Union

import pandas as pd
from prompt_toolkit.utils import is_windows

from hummingbot.client.command.gateway_command import GatewayCommand
from hummingbot.client.config.config_helpers import (
    ClientConfigAdapter,
    missing_required_configs_legacy,
    save_to_yml,
    save_to_yml_legacy,
)
from hummingbot.client.config.config_validators import validate_bool, validate_decimal
from hummingbot.client.config.config_var import ConfigVar
from hummingbot.client.config.security import Security
from hummingbot.client.config.strategy_config_data_types import BaseTradingStrategyConfigMap
from hummingbot.client.settings import CLIENT_CONFIG_PATH, STRATEGIES_CONF_DIR_PATH
from hummingbot.client.ui.interface_utils import format_df_for_printout
from hummingbot.client.ui.style import load_style
from hummingbot.connector.utils import split_hb_trading_pair
from hummingbot.core.utils import map_df_to_str
from hummingbot.core.utils.async_utils import safe_ensure_future
from hummingbot.model.inventory_cost import InventoryCost
from hummingbot.strategy.perpetual_market_making import PerpetualMarketMakingStrategy
from hummingbot.strategy.pure_market_making import PureMarketMakingStrategy
from hummingbot.user.user_balances import UserBalances

if TYPE_CHECKING:
    from hummingbot.client.hummingbot_application import HummingbotApplication  # noqa: F401

no_restart_pmm_keys_in_percentage = ["bid_spread", "ask_spread", "order_level_spread", "inventory_target_base_pct"]
no_restart_pmm_keys = ["order_amount",
                       "order_levels",
                       "filled_order_delay",
                       "inventory_skew_enabled",
                       "inventory_range_multiplier",
                       "price_ceiling",
                       "price_floor",
                       "moving_price_band_enabled",
                       "price_ceiling_pct",
                       "price_floor_pct",
                       "price_band_refresh_time"
                       "order_optimization_enabled",
                       "bid_order_optimization_depth",
                       "ask_order_optimization_depth"
                       ]
client_configs_to_display = ["autofill_import",
                             "kill_switch_mode",
                             "kill_switch_rate",
                             "mqtt_bridge",
                             "mqtt_host",
                             "mqtt_port",
                             "mqtt_namespace",
                             "mqtt_username",
                             "mqtt_password",
                             "mqtt_ssl",
                             "mqtt_logger",
                             "mqtt_notifier",
                             "mqtt_commands",
                             "mqtt_events",
                             "mqtt_external_events",
                             "mqtt_autostart",
                             "instance_id",
                             "send_error_logs",
                             "ethereum_chain_name",
                             "gateway",
                             "gateway_api_host",
                             "gateway_api_port",
                             "rate_oracle_source",
                             "extra_tokens",
                             "fetch_pairs_from_all_exchanges",
                             "global_token",
                             "global_token_name",
                             "global_token_symbol",
                             "rate_limits_share_pct",
                             "commands_timeout",
                             "create_command_timeout",
                             "other_commands_timeout",
                             "tables_format",
                             "tick_size",
                             "market_data_collection",
                             "market_data_collection_enabled",
                             "market_data_collection_interval",
                             "market_data_collection_depth",
                             ]
color_settings_to_display = ["top_pane",
                             "bottom_pane",
                             "output_pane",
                             "input_pane",
                             "logs_pane",
                             "terminal_primary"]
columns = ["Key", "Value"]


class ConfigCommand:
    def config(self,  # type: HummingbotApplication
               key: str = None,
               value: str = None):
        self.app.clear_input()
        if key is None:
            self.list_configs()
            return
        else:
            if key not in self.configurable_keys():
                self.notify("Invalid key, please choose from the list.")
                return
            safe_ensure_future(self._config_single_key(key, value), loop=self.ev_loop)

    def list_configs(self,  # type: HummingbotApplication
                     ):
        self.list_client_configs()
        self.list_strategy_configs()

    def list_client_configs(
            self,  # type: HummingbotApplication
    ):
        data = self.build_model_df_data(self.client_config_map, to_print=client_configs_to_display)
        df = map_df_to_str(pd.DataFrame(data=data, columns=columns))
        self.notify("\nGlobal Configurations:")
        lines = ["    " + line for line in format_df_for_printout(
            df,
            table_format=self.client_config_map.tables_format,
            max_col_width=50).split("\n")]
        self.notify("\n".join(lines))

        data = self.build_model_df_data(self.client_config_map, to_print=color_settings_to_display)
        df = map_df_to_str(pd.DataFrame(data=data, columns=columns))
        self.notify("\nColor Settings:")
        lines = ["    " + line for line in format_df_for_printout(
            df,
            table_format=self.client_config_map.tables_format,
            max_col_width=50).split("\n")]
        self.notify("\n".join(lines))

    def list_strategy_configs(
            self,  # type: HummingbotApplication
    ):
        if self.strategy_name is not None:
            config_map = self.strategy_config_map
            data = self.build_df_data_from_config_map(config_map)
            df = map_df_to_str(pd.DataFrame(data=data, columns=columns))
            self.notify("\nStrategy Configurations:")
            lines = ["    " + line for line in format_df_for_printout(
                df,
                table_format=self.client_config_map.tables_format,
                max_col_width=50).split("\n")]
            self.notify("\n".join(lines))

    def build_df_data_from_config_map(
            self,  # type: HummingbotApplication
            config_map: Union[ClientConfigAdapter, Dict[str, ConfigVar]]
    ) -> List[Tuple[str, Any]]:
        if isinstance(config_map, ClientConfigAdapter):
            data = self.build_model_df_data(config_map)
        else:  # legacy
            data = [[cv.printable_key or cv.key, cv.value] for cv in self.strategy_config_map.values() if
                    not cv.is_secure]
        return data

    @staticmethod
    def build_model_df_data(
            config_map: ClientConfigAdapter, to_print: Optional[List[str]] = None
    ) -> List[Tuple[str, Any]]:
        model_data = []
        for traversal_item in config_map.traverse():
            if to_print is not None and traversal_item.attr not in to_print:
                continue
            attr_printout = (
                "  " * (traversal_item.depth - 1)
                + (u"\u221F " if not is_windows() else "  ")
                + traversal_item.attr
            ) if traversal_item.depth else traversal_item.attr
            model_data.append((attr_printout, traversal_item.printable_value))
        return model_data

    def configurable_keys(self,  # type: HummingbotApplication
                          ) -> List[str]:
        """
        Returns a list of configurable keys - using config command, excluding exchanges api keys
        as they are set from connect command.
        """
        keys = [
            traversal_item.config_path
            for traversal_item in self.client_config_map.traverse()
            if (traversal_item.client_field_data is not None and traversal_item.client_field_data.prompt is not None)
        ]
        if self.strategy_config_map is not None:
            if isinstance(self.strategy_config_map, ClientConfigAdapter):
                keys.extend([
                    traversal_item.config_path
                    for traversal_item in self.strategy_config_map.traverse()
                    if (traversal_item.client_field_data is not None
                        and traversal_item.client_field_data.prompt is not None)
                ])
            else:  # legacy
                keys.extend(
                    [c.key for c in self.strategy_config_map.values() if c.prompt is not None and c.key != 'strategy'])
        return keys

    async def check_password(self,  # type: HummingbotApplication
                             ):
        password = await self.app.prompt(prompt="Enter your password >>> ", is_password=True)
        if password != Security.secrets_manager.password.get_secret_value():
            self.notify("Invalid password, please try again.")
            return False
        else:
            return True

    # Make this function static so unit testing can be performed.
    @staticmethod
    def update_running_mm(mm_strategy, key: str, new_value: Any):
        if key in no_restart_pmm_keys_in_percentage:
            setattr(mm_strategy, key, new_value / Decimal("100"))
            return True
        elif key in no_restart_pmm_keys:
            setattr(mm_strategy, key, new_value)
            return True
        return False

    async def _config_single_key(self,  # type: HummingbotApplication
                                 key: str,
                                 input_value):
        """
        Configure a single variable only.
        Prompt the user to finish all configurations if there are remaining empty configs at the end.
        """

        self.placeholder_mode = True
        self.app.hide_input = True

        try:
            if (
                    not isinstance(self.strategy_config_map, (type(None), ClientConfigAdapter))
                    and key in self.strategy_config_map
            ):
                await self._config_single_key_legacy(key, input_value)
            else:
                client_config_key = key in self.client_config_map.config_paths()
                if client_config_key:
                    config_map = self.client_config_map
                    file_path = CLIENT_CONFIG_PATH
                elif self.strategy is not None:
                    self.notify("Configuring the strategy while it is running is not currently supported.")
                    return
                else:
                    config_map = self.strategy_config_map
                    if self.strategy_file_name is not None:
                        file_path = STRATEGIES_CONF_DIR_PATH / self.strategy_file_name
                    else:
                        self.notify("Strategy file name is not configured.")
                        return

                if input_value is None:
                    self.notify("Please follow the prompt to complete configurations: ")
                if key == "inventory_target_base_pct":
                    await self.asset_ratio_maintenance_prompt(config_map, input_value)
                elif key == "inventory_price":
                    await self.inventory_price_prompt(config_map, input_value)
                else:
                    await self.prompt_a_config(config_map, key, input_value, assign_default=False)
                if self.app.to_stop_config:
                    self.app.to_stop_config = False
                    return
                save_to_yml(file_path, config_map)
                self.notify("\nNew configuration saved.")
                if client_config_key:
                    self.list_client_configs()
                else:
                    self.list_strategy_configs()
                self.app.style = load_style(self.client_config_map)
        except asyncio.TimeoutError:
            self.logger().error("Prompt timeout")
        except Exception as err:
            self.logger().error(str(err), exc_info=True)
        finally:
            self.app.hide_input = False
            self.placeholder_mode = False
            self.app.change_prompt(prompt=">>> ")

    async def _config_single_key_legacy(
            self,  # type: HummingbotApplication
            key: str,
            input_value: Any,
    ):  # pragma: no cover
        config_var, config_map, file_path = None, None, None
        if self.strategy_config_map is not None and key in self.strategy_config_map:
            config_map = self.strategy_config_map
            file_path = STRATEGIES_CONF_DIR_PATH / self.strategy_file_name
        config_var = config_map[key]
        if config_var.key == "strategy":
            self.notify("You cannot change the strategy of a loaded configuration.")
            self.notify("Please use 'import xxx.yml' or 'create' to configure the intended strategy")
            return
        if input_value is None:
            self.notify("Please follow the prompt to complete configurations: ")
        if config_var.key == "inventory_target_base_pct":
            await self.asset_ratio_maintenance_prompt_legacy(config_map, input_value)
        elif config_var.key == "inventory_price":
            await self.inventory_price_prompt_legacy(config_map, input_value)
        else:
            await self.prompt_a_config_legacy(config_var, input_value=input_value, assign_default=False)
        if self.app.to_stop_config:
            self.app.to_stop_config = False
            return
        missings = missing_required_configs_legacy(config_map)
        if missings:
            self.notify("\nThere are other configuration required, please follow the prompt to complete them.")
        missings = await self._prompt_missing_configs(config_map)
        save_to_yml_legacy(str(file_path), config_map)
        self.notify("\nNew configuration saved:")
        self.notify(f"{key}: {str(config_var.value)}")
        self.app.app.style = load_style(self.client_config_map)
        for config in missings:
            self.notify(f"{config.key}: {str(config.value)}")
        if (
                isinstance(self.strategy, PureMarketMakingStrategy) or
                isinstance(self.strategy, PerpetualMarketMakingStrategy)
        ):
            updated = ConfigCommand.update_running_mm(self.strategy, key, config_var.value)
            if updated:
                self.notify(f"\nThe current {self.strategy_name} strategy has been updated "
                            f"to reflect the new configuration.")

    async def _prompt_missing_configs(self,  # type: HummingbotApplication
                                      config_map):
        missings = missing_required_configs_legacy(config_map)
        for config in missings:
            await self.prompt_a_config_legacy(config)
            if self.app.to_stop_config:
                self.app.to_stop_config = False
                return
        if missing_required_configs_legacy(config_map):
            return missings + (await self._prompt_missing_configs(config_map))
        return missings

    async def asset_ratio_maintenance_prompt(
            self,  # type: HummingbotApplication
            config_map: BaseTradingStrategyConfigMap,
            input_value: Any = None,
    ):  # pragma: no cover
        if input_value:
            config_map.inventory_target_base_pct = input_value
        else:
            exchange = config_map.exchange
            market = config_map.market
            base, quote = split_hb_trading_pair(market)
            if UserBalances.instance().is_gateway_market(exchange):
                balances = await GatewayCommand.balance(self, exchange, config_map, base, quote)
            else:
                balances = await UserBalances.instance().balances(exchange, config_map, base, quote)
            if balances is None:
                return
            base_ratio = await UserBalances.base_amount_ratio(exchange, market, balances)
            if base_ratio is None:
                return
            base_ratio = round(base_ratio, 3)
            quote_ratio = 1 - base_ratio

            cvar = ConfigVar(key="temp_config",
                             prompt=f"On {exchange}, you have {balances.get(base, 0):.4f} {base} and "
                                    f"{balances.get(quote, 0):.4f} {quote}. By market value, "
                                    f"your current inventory split is {base_ratio:.1%} {base} "
                                    f"and {quote_ratio:.1%} {quote}."
                                    f" Would you like to keep this ratio? (Yes/No) >>> ",
                             required_if=lambda: True,
                             type_str="bool",
                             validator=validate_bool)
            await self.prompt_a_config_legacy(cvar)
            if cvar.value:
                config_map.inventory_target_base_pct = round(base_ratio * Decimal('100'), 1)
            elif self.app.to_stop_config:
                self.app.to_stop_config = False
            else:
                await self.prompt_a_config(config_map, config="inventory_target_base_pct")

    async def asset_ratio_maintenance_prompt_legacy(
            self,  # type: HummingbotApplication
            config_map,
            input_value=None,
    ):
        if input_value:
            config_map['inventory_target_base_pct'].value = Decimal(input_value)
        else:
            exchange = config_map['exchange'].value
            market = config_map["market"].value
            base, quote = market.split("-")
            if UserBalances.instance().is_gateway_market(exchange):
                balances = await GatewayCommand.balance(self, exchange, config_map, base, quote)
            else:
                balances = await UserBalances.instance().balances(exchange, config_map, base, quote)
            if balances is None:
                return
            base_ratio = await UserBalances.base_amount_ratio(exchange, market, balances)
            if base_ratio is None:
                return
            base_ratio = round(base_ratio, 3)
            quote_ratio = 1 - base_ratio
            base, quote = config_map["market"].value.split("-")

            cvar = ConfigVar(key="temp_config",
                             prompt=f"On {exchange}, you have {balances.get(base, 0):.4f} {base} and "
                                    f"{balances.get(quote, 0):.4f} {quote}. By market value, "
                                    f"your current inventory split is {base_ratio:.1%} {base} "
                                    f"and {quote_ratio:.1%} {quote}."
                                    f" Would you like to keep this ratio? (Yes/No) >>> ",
                             required_if=lambda: True,
                             type_str="bool",
                             validator=validate_bool)
            await self.prompt_a_config_legacy(cvar)
            if cvar.value:
                config_map['inventory_target_base_pct'].value = round(base_ratio * Decimal('100'), 1)
            else:
                if self.app.to_stop_config:
                    self.app.to_stop_config = False
                    return
                await self.prompt_a_config_legacy(config_map["inventory_target_base_pct"])

    async def inventory_price_prompt(
            self,  # type: HummingbotApplication
            model: BaseTradingStrategyConfigMap,
            input_value=None,
    ):
        """
        Not currently used.
        """
        raise NotImplementedError

    async def inventory_price_prompt_legacy(
            self,  # type: HummingbotApplication
            config_map,
            input_value=None,
    ):
        key = "inventory_price"
        if input_value:
            config_map[key].value = Decimal(input_value)
        else:
            exchange = config_map["exchange"].value
            market = config_map["market"].value
            base_asset, quote_asset = market.split("-")

            if exchange.endswith("paper_trade"):
                balances = self.client_config_map.paper_trade.paper_trade_account_balance
            elif UserBalances.instance().is_gateway_market(exchange):
                balances = await GatewayCommand.balance(self, exchange, config_map, base_asset, quote_asset)
            else:
                balances = await UserBalances.instance().balances(
                    exchange, base_asset, quote_asset
                )
            if balances.get(base_asset) is None:
                return

            cvar = ConfigVar(
                key="temp_config",
                prompt=f"On {exchange}, you have {balances[base_asset]:.4f} {base_asset}. "
                       f"What was the price for this amount in {quote_asset}?  >>> ",
                required_if=lambda: True,
                type_str="decimal",
                validator=lambda v: validate_decimal(
                    v, min_value=Decimal("0"), inclusive=True
                ),
            )
            await self.prompt_a_config_legacy(cvar)
            config_map[key].value = cvar.value

            try:
                quote_volume = balances[base_asset] * cvar.value
            except TypeError:
                # TypeError: unsupported operand type(s) for *: 'decimal.Decimal' and 'NoneType' - bad input / no input
                self.notify("Inventory price not updated due to bad input")
                return

            with self.trade_fill_db.get_new_session() as session:
                with session.begin():
                    InventoryCost.add_volume(
                        session,
                        base_asset=base_asset,
                        quote_asset=quote_asset,
                        base_volume=balances[base_asset],
                        quote_volume=quote_volume,
                        overwrite=True,
                    )



from typing import List

from hummingbot.client.config.config_helpers import ClientConfigAdapter, get_connector_class
from hummingbot.client.settings import AllConnectorSettings
from hummingbot.connector.exchange.paper_trade.paper_trade_exchange import PaperTradeExchange
from hummingbot.core.data_type.order_book_tracker import OrderBookTracker


def get_order_book_tracker(connector_name: str, trading_pairs: List[str]) -> OrderBookTracker:
    conn_setting = AllConnectorSettings.get_connector_settings()[connector_name]
    try:
        connector_instance = conn_setting.non_trading_connector_instance_with_default_configuration(
            trading_pairs=trading_pairs)
        return connector_instance.order_book_tracker
    except Exception as exception:
        raise Exception(f"Connector {connector_name} OrderBookTracker class not found ({exception})")


def create_paper_trade_market(exchange_name: str, client_config_map: ClientConfigAdapter, trading_pairs: List[str]):
    tracker = get_order_book_tracker(connector_name=exchange_name, trading_pairs=trading_pairs)
    return PaperTradeExchange(client_config_map,
                              tracker,
                              get_connector_class(exchange_name),
                              exchange_name=exchange_name)



from decimal import Decimal

from hummingbot.strategy.market_trading_pair_tuple import MarketTradingPairTuple

s_decimal_nan = Decimal("NaN")
s_decimal_0 = Decimal("0")


class ArbProposalSide:
    """
    An arbitrage proposal side which contains info needed for order submission.
    """
    def __init__(self,
                 market_info: MarketTradingPairTuple,
                 is_buy: bool,
                 order_price: Decimal
                 ):
        """
        :param market_info: The market where to submit the order
        :param is_buy: True if buy order
        :param order_price: The price required for order submission, this could differ from the quote price
        """
        self.market_info: MarketTradingPairTuple = market_info
        self.is_buy: bool = is_buy
        self.order_price: Decimal = order_price

    def __repr__(self):
        side = "Buy" if self.is_buy else "Sell"
        base, quote = self.market_info.trading_pair.split("-")
        return f"{self.market_info.market.display_name.capitalize()}: {side} {base}" \
               f" at {self.order_price} {quote}."


class ArbProposal:
    """
    An arbitrage proposal which contains 2 sides of the proposal - one on spot market and one on perpetual market.
    """
    def __init__(self,
                 spot_side: ArbProposalSide,
                 perp_side: ArbProposalSide,
                 order_amount: Decimal):
        """
        Creates ArbProposal
        :param spot_side: An ArbProposalSide on spot market
        :param perp_side: An ArbProposalSide on perpetual market
        :param order_amount: An order amount for both spot and perpetual market
        """
        if spot_side.is_buy == perp_side.is_buy:
            raise Exception("Spot and perpetual arb proposal cannot be on the same side.")
        self.spot_side: ArbProposalSide = spot_side
        self.perp_side: ArbProposalSide = perp_side
        self.order_amount: Decimal = order_amount

    def profit_pct(self) -> Decimal:
        """
        Calculates and returns arbitrage profit (in percentage value).
        """
        buy_price = self.spot_side.order_price if self.spot_side.is_buy else self.perp_side.order_price
        sell_price = self.spot_side.order_price if not self.spot_side.is_buy else self.perp_side.order_price
        if sell_price and buy_price:
            return (sell_price - buy_price) / buy_price
        return s_decimal_0

    def __repr__(self):
        return f"Spot: {self.spot_side}\nPerpetual: {self.perp_side}\nOrder amount: {self.order_amount}\n" \
               f"Profit: {self.profit_pct():.2%}"



import asyncio
import math
from decimal import Decimal
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Tuple

from bidict import bidict

from hummingbot.connector.constants import s_decimal_NaN
from hummingbot.connector.exchange.bitmart import (
    bitmart_constants as CONSTANTS,
    bitmart_utils,
    bitmart_web_utils as web_utils,
)
from hummingbot.connector.exchange.bitmart.bitmart_api_order_book_data_source import BitmartAPIOrderBookDataSource
from hummingbot.connector.exchange.bitmart.bitmart_api_user_stream_data_source import BitmartAPIUserStreamDataSource
from hummingbot.connector.exchange.bitmart.bitmart_auth import BitmartAuth
from hummingbot.connector.exchange_py_base import ExchangePyBase
from hummingbot.connector.trading_rule import TradingRule
from hummingbot.connector.utils import combine_to_hb_trading_pair
from hummingbot.core.data_type.common import OrderType, TradeType
from hummingbot.core.data_type.in_flight_order import InFlightOrder, OrderState, OrderUpdate, TradeUpdate
from hummingbot.core.data_type.order_book_tracker_data_source import OrderBookTrackerDataSource
from hummingbot.core.data_type.trade_fee import AddedToCostTradeFee, TokenAmount, TradeFeeBase
from hummingbot.core.data_type.user_stream_tracker_data_source import UserStreamTrackerDataSource
from hummingbot.core.web_assistant.web_assistants_factory import WebAssistantsFactory

if TYPE_CHECKING:
    from hummingbot.client.config.config_helpers import ClientConfigAdapter


class BitmartExchange(ExchangePyBase):
    """
    BitmartExchange connects with BitMart exchange and provides order book pricing, user account tracking and
    trading functionality.
    """
    API_CALL_TIMEOUT = 10.0
    POLL_INTERVAL = 1.0
    UPDATE_ORDER_STATUS_MIN_INTERVAL = 10.0
    UPDATE_TRADE_STATUS_MIN_INTERVAL = 10.0

    web_utils = web_utils

    def __init__(self,
                 client_config_map: "ClientConfigAdapter",
                 bitmart_api_key: str,
                 bitmart_secret_key: str,
                 bitmart_memo: str,
                 trading_pairs: Optional[List[str]] = None,
                 trading_required: bool = True,
                 ):
        """
        :param bitmart_api_key: The API key to connect to private BitMart APIs.
        :param bitmart_secret_key: The API secret.
        :param trading_pairs: The market trading pairs which to track order book data.
        :param trading_required: Whether actual trading is needed.
        """
        self._api_key: str = bitmart_api_key
        self._secret_key: str = bitmart_secret_key
        self._memo: str = bitmart_memo
        self._trading_required = trading_required
        self._trading_pairs = trading_pairs

        super().__init__(client_config_map)
        self.real_time_balance_update = False

    @property
    def authenticator(self):
        return BitmartAuth(
            api_key=self._api_key,
            secret_key=self._secret_key,
            memo=self._memo,
            time_provider=self._time_synchronizer)

    @property
    def name(self) -> str:
        return "bitmart"

    @property
    def rate_limits_rules(self):
        return CONSTANTS.RATE_LIMITS

    @property
    def domain(self):
        return CONSTANTS.DEFAULT_DOMAIN

    @property
    def client_order_id_max_length(self):
        return CONSTANTS.MAX_ORDER_ID_LEN

    @property
    def client_order_id_prefix(self):
        return CONSTANTS.HBOT_ORDER_ID_PREFIX

    @property
    def trading_rules_request_path(self):
        return CONSTANTS.GET_TRADING_RULES_PATH_URL

    @property
    def trading_pairs_request_path(self):
        return CONSTANTS.GET_TRADING_RULES_PATH_URL

    @property
    def check_network_request_path(self):
        return CONSTANTS.CHECK_NETWORK_PATH_URL

    @property
    def trading_pairs(self):
        return self._trading_pairs

    @property
    def is_cancel_request_in_exchange_synchronous(self) -> bool:
        return True

    @property
    def is_trading_required(self) -> bool:
        return self._trading_required

    def supported_order_types(self) -> List[OrderType]:
        """
        :return a list of OrderType supported by this connector.
        """
        return [OrderType.LIMIT, OrderType.MARKET, OrderType.LIMIT_MAKER]

    def _is_request_exception_related_to_time_synchronizer(self, request_exception: Exception):
        error_description = str(request_exception)
        is_time_synchronizer_related = ("Header X-BM-TIMESTAMP" in error_description
                                        and ("30007" in error_description or "30008" in error_description))
        return is_time_synchronizer_related

    def _is_order_not_found_during_status_update_error(self, status_update_exception: Exception) -> bool:
        # TODO: implement this method correctly for the connector
        # The default implementation was added when the functionality to detect not found orders was introduced in the
        # ExchangePyBase class. Also fix the unit test test_lost_order_removed_if_not_found_during_order_status_update
        # when replacing the dummy implementation
        return False

    def _is_order_not_found_during_cancelation_error(self, cancelation_exception: Exception) -> bool:
        # TODO: implement this method correctly for the connector
        # The default implementation was added when the functionality to detect not found orders was introduced in the
        # ExchangePyBase class. Also fix the unit test test_cancel_order_not_found_in_the_exchange when replacing the
        # dummy implementation
        return False

    def _create_web_assistants_factory(self) -> WebAssistantsFactory:
        return web_utils.build_api_factory(
            throttler=self._throttler,
            time_synchronizer=self._time_synchronizer,
            auth=self._auth)

    def _create_order_book_data_source(self) -> OrderBookTrackerDataSource:
        return BitmartAPIOrderBookDataSource(
            trading_pairs=self._trading_pairs,
            connector=self,
            api_factory=self._web_assistants_factory)

    def _create_user_stream_data_source(self) -> UserStreamTrackerDataSource:
        return BitmartAPIUserStreamDataSource(
            auth=self._auth,
            trading_pairs=self._trading_pairs,
            connector=self,
            api_factory=self._web_assistants_factory,
        )

    def _get_fee(self,
                 base_currency: str,
                 quote_currency: str,
                 order_type: OrderType,
                 order_side: TradeType,
                 amount: Decimal,
                 price: Decimal = s_decimal_NaN,
                 is_maker: Optional[bool] = None) -> AddedToCostTradeFee:
        """
        To get trading fee, this function is simplified by using fee override configuration. Most parameters to this
        function are ignore except order_type. Use OrderType.LIMIT_MAKER to specify you want trading fee for
        maker order.
        """
        is_maker = order_type is OrderType.LIMIT_MAKER
        return AddedToCostTradeFee(percent=self.estimate_fee_pct(is_maker))

    async def _place_order(self,
                           order_id: str,
                           trading_pair: str,
                           amount: Decimal,
                           trade_type: TradeType,
                           order_type: OrderType,
                           price: Decimal,
                           **kwargs) -> Tuple[str, float]:

        if order_type is OrderType.MARKET:
            price = await self._get_last_traded_price(trading_pair)
        notionalValue: Decimal = (amount * Decimal(price))
        api_params = {"symbol": await self.exchange_symbol_associated_to_pair(trading_pair),
                      "side": trade_type.name.lower(),
                      "type": order_type.name.lower(),
                      "size": f"{amount:f}",
                      "price": f"{price:f}",
                      "client_order_id": order_id,
                      "notional": f"{notionalValue:f}",
                      }
        order_result = await self._api_post(
            path_url=CONSTANTS.CREATE_ORDER_PATH_URL,
            data=api_params,
            is_auth_required=True)
        exchange_order_id = str(order_result["data"]["order_id"])

        return exchange_order_id, self.current_timestamp

    async def _place_cancel(self, order_id: str, tracked_order: InFlightOrder):
        symbol = await self.exchange_symbol_associated_to_pair(trading_pair=tracked_order.trading_pair)
        api_params = {
            "symbol": symbol,
            "client_order_id": order_id,
        }
        cancel_result = await self._api_post(
            path_url=CONSTANTS.CANCEL_ORDER_PATH_URL,
            data=api_params,
            is_auth_required=True)
        # await cancel_result.get("data", {}).get("result", False)
        return bool(cancel_result["data"]["result"])

    async def _format_trading_rules(self, symbols_details: Dict[str, Any]) -> List[TradingRule]:
        """
        Converts json API response into a dictionary of trading rules.
        :param symbols_details: The json API response
        :return A dictionary of trading rules.
        Response Example:
        {
            "code": 1000,
            "trace":"886fb6ae-456b-4654-b4e0-d681ac05cea1",
            "message": "OK",
            "data": {
                "symbols": [
                    {
                        "symbol":"GXC_BTC",
                         "symbol_id":1024,
                         "base_currency":"GXC",
                         "quote_currency":"BTC",
                         "quote_increment":"1.00000000",
                         "base_min_size":"1.00000000",
                         "price_min_precision":6,
                         "price_max_precision":8,
                         "expiration":"NA",
                         "min_buy_amount":"0.00010000",
                         "min_sell_amount":"0.00010000"
                    },
                    ...
                ]
            }
        }
        """
        result = []
        for rule in symbols_details["data"]["symbols"]:
            if bitmart_utils.is_exchange_information_valid(rule):
                try:
                    trading_pair = await self.trading_pair_associated_to_exchange_symbol(rule["symbol"])
                    price_decimals = Decimal(str(rule["price_max_precision"]))
                    # E.g. a price decimal of 2 means 0.01 incremental.
                    price_step = Decimal("1") / Decimal(str(math.pow(10, price_decimals)))
                    result.append(TradingRule(trading_pair=trading_pair,
                                              min_order_size=Decimal(str(rule["base_min_size"])),
                                              min_order_value=Decimal(str(rule["min_buy_amount"])),
                                              min_base_amount_increment=Decimal(str(rule["base_min_size"])),
                                              min_price_increment=price_step))
                except KeyError:
                    # Ignore results for which their symbols is not tracked by the connector
                    continue
                except Exception:
                    self.logger().exception(f"Error parsing the trading pair rule {rule}. Skipping.")
        return result

    async def _update_trading_fees(self):
        """
        Update fees information from the exchange
        """
        pass

    async def _update_balances(self):
        """
        Calls REST API to update total and available balances.
        """
        local_asset_names = set(self._account_balances.keys())
        remote_asset_names = set()
        account_info = await self._api_get(
            path_url=CONSTANTS.GET_ACCOUNT_SUMMARY_PATH_URL,
            is_auth_required=True)
        for account in account_info["data"]["wallet"]:
            asset_name = account["id"]
            self._account_available_balances[asset_name] = Decimal(str(account["available"]))
            self._account_balances[asset_name] = Decimal(str(account["available"])) + Decimal(str(account["frozen"]))
            remote_asset_names.add(asset_name)

        asset_names_to_remove = local_asset_names.difference(remote_asset_names)
        for asset_name in asset_names_to_remove:
            del self._account_available_balances[asset_name]
            del self._account_balances[asset_name]

    async def _request_order_update(self, order: InFlightOrder) -> Dict[str, Any]:
        return await self._api_post(
            path_url=CONSTANTS.GET_ORDER_DETAIL_PATH_URL,
            data={"orderId": order.exchange_order_id},
            is_auth_required=True)

    async def _request_order_fills(self, order: InFlightOrder) -> Dict[str, Any]:
        return await self._api_post(
            path_url=CONSTANTS.GET_TRADE_DETAIL_PATH_URL,
            data={"orderId": order.exchange_order_id},
            is_auth_required=True)

    async def _all_trade_updates_for_order(self, order: InFlightOrder) -> List[TradeUpdate]:
        trade_updates = []

        try:
            if order.exchange_order_id is not None:
                all_fills_response = await self._request_order_fills(order=order)
                updates = self._create_order_fill_updates(order=order, fill_update=all_fills_response)
                trade_updates.extend(updates)
        except asyncio.CancelledError:
            raise
        except Exception as ex:
            is_error_caused_by_unexistent_order = '"code":50005' in str(ex)
            if not is_error_caused_by_unexistent_order:
                raise

        return trade_updates

    async def _request_order_status(self, tracked_order: InFlightOrder) -> OrderUpdate:
        updated_order_data = await self._request_order_update(order=tracked_order)

        order_update = self._create_order_update(order=tracked_order, order_update=updated_order_data)
        return order_update

    def _create_order_fill_updates(self, order: InFlightOrder, fill_update: Dict[str, Any]) -> List[TradeUpdate]:
        updates = []
        fills_data = fill_update["data"]

        for fill_data in fills_data:
            fee = TradeFeeBase.new_spot_fee(
                fee_schema=self.trade_fee_schema(),
                trade_type=order.trade_type,
                percent_token=fill_data["feeCoinName"],
                flat_fees=[TokenAmount(amount=Decimal(fill_data["fee"]), token=fill_data["feeCoinName"])]
            )
            trade_update = TradeUpdate(
                trade_id=str(fill_data["tradeId"]),
                client_order_id=order.client_order_id,
                exchange_order_id=str(fill_data["orderId"]),
                trading_pair=order.trading_pair,
                fee=fee,
                fill_base_amount=Decimal(fill_data["size"]),
                fill_quote_amount=Decimal(fill_data["notional"]),
                fill_price=Decimal(fill_data["price"]),
                fill_timestamp=int(fill_data["createTime"]) * 1e-3,
            )
            updates.append(trade_update)

        return updates

    def _create_order_update(self, order: InFlightOrder, order_update: Dict[str, Any]) -> OrderUpdate:
        order_data = order_update["data"]
        new_state = CONSTANTS.ORDER_STATE[order_data["state"]]
        # This is a workaround to account for a MARKET BUY order reporting the state as "partially cancelled"
        # Bitmart reports this state for a successfully filled MARKET BUY order which is confusing.
        if order_data["state"] == "partially_canceled" and order_data["type"] == "market" and order_data["side"] == "buy":
            new_state = OrderState.FILLED
        update = OrderUpdate(
            client_order_id=order.client_order_id,
            exchange_order_id=str(order_data["orderId"]),
            trading_pair=order.trading_pair,
            update_timestamp=self.current_timestamp,
            new_state=new_state,
        )
        return update

    async def _user_stream_event_listener(self):
        async for event_message in self._iter_user_event_queue():
            try:
                event_type = event_message.get("table")
                execution_data = event_message.get("data", [])

                # Refer to https://developer-pro.bitmart.com/en/spot/#private-order-progress
                if event_type == CONSTANTS.PRIVATE_ORDER_PROGRESS_CHANNEL_NAME:
                    for each_event in execution_data:
                        try:
                            client_order_id: Optional[str] = each_event.get("client_order_id")
                            fillable_order = self._order_tracker.all_fillable_orders.get(client_order_id)
                            updatable_order = self._order_tracker.all_updatable_orders.get(client_order_id)

                            new_state = CONSTANTS.ORDER_STATE[each_event["order_state"]]
                            # This is a workaround to account for a MARKET BUY order reporting the state as "partially cancelled"
                            # Bitmart reports this state for a successfully filled MARKET BUY order which is confusing.
                            if each_event["order_state"] == "partially_canceled" and each_event["type"] == "market" and each_event["side"] == "buy":
                                new_state = CONSTANTS.ORDER_STATE["filled"]
                            event_timestamp = int(each_event["ms_t"]) * 1e-3

                            if fillable_order is not None:
                                is_fill_candidate_by_state = new_state in [OrderState.PARTIALLY_FILLED,
                                                                           OrderState.FILLED]
                                is_fill_candidate_by_amount = fillable_order.executed_amount_base < Decimal(
                                    each_event["filled_size"])
                                if is_fill_candidate_by_state and is_fill_candidate_by_amount:
                                    try:
                                        trade_fills: Dict[str, Any] = await self._request_order_fills(fillable_order)
                                        trade_updates = self._create_order_fill_updates(
                                            order=fillable_order,
                                            fill_update=trade_fills)
                                        for trade_update in trade_updates:
                                            self._order_tracker.process_trade_update(trade_update)
                                    except asyncio.CancelledError:
                                        raise
                                    except Exception:
                                        self.logger().exception("Unexpected error requesting order fills for "
                                                                f"{fillable_order.client_order_id}")
                            if updatable_order is not None:
                                order_update = OrderUpdate(
                                    trading_pair=updatable_order.trading_pair,
                                    update_timestamp=event_timestamp,
                                    new_state=new_state,
                                    client_order_id=client_order_id,
                                    exchange_order_id=each_event["order_id"],
                                )
                                self._order_tracker.process_order_update(order_update=order_update)

                        except asyncio.CancelledError:
                            raise
                        except Exception:
                            self.logger().exception("Unexpected error in user stream listener loop.")
            except asyncio.CancelledError:
                raise
            except Exception:
                self.logger().exception("Unexpected error in user stream listener loop.")

    def _initialize_trading_pair_symbols_from_exchange_info(self, exchange_info: Dict[str, Any]):
        mapping = bidict()
        for symbol_data in filter(bitmart_utils.is_exchange_information_valid, exchange_info["data"]["symbols"]):
            mapping[symbol_data["symbol"]] = combine_to_hb_trading_pair(base=symbol_data["base_currency"],
                                                                        quote=symbol_data["quote_currency"])
        self._set_trading_pair_symbol_map(mapping)

    async def _get_last_traded_price(self, trading_pair: str) -> float:
        params = {
            "symbol": await self.exchange_symbol_associated_to_pair(trading_pair=trading_pair)
        }

        resp_json = await self._api_get(
            path_url=CONSTANTS.GET_LAST_TRADING_PRICES_PATH_URL,
            params=params
        )

        return float(resp_json["data"]["last"])



#!/usr/bin/env python

from typing import Callable

from hummingbot.core.event.event_listener import EventListener
from hummingbot.core.pubsub import PubSub


class EventForwarder(EventListener):
    def __init__(self, to_function: Callable[[any], None]):
        super().__init__()
        self._to_function: Callable[[any], None] = to_function

    def __call__(self, arg: any):
        self._to_function(arg)


class SourceInfoEventForwarder(EventListener):
    def __init__(self, to_function: Callable[[int, PubSub, any], None]):
        super().__init__()
        self._to_function: Callable[[int, PubSub, any], None] = to_function

    def __call__(self, arg: any):
        self._to_function(self.current_event_tag, self.current_event_caller, arg)



import base64
import hashlib
import hmac
from collections import OrderedDict
from typing import Any, Dict
from urllib.parse import urlencode

from hummingbot.connector.exchange.kucoin import kucoin_constants as CONSTANTS
from hummingbot.connector.time_synchronizer import TimeSynchronizer
from hummingbot.core.web_assistant.auth import AuthBase
from hummingbot.core.web_assistant.connections.data_types import RESTRequest, WSRequest


class KucoinAuth(AuthBase):
    def __init__(self, api_key: str, passphrase: str, secret_key: str, time_provider: TimeSynchronizer):
        self.api_key: str = api_key
        self.passphrase: str = passphrase
        self.secret_key: str = secret_key
        self.time_provider = time_provider

    @staticmethod
    def keysort(dictionary: Dict[str, str]) -> Dict[str, str]:
        return OrderedDict(sorted(dictionary.items(), key=lambda t: t[0]))

    async def rest_authenticate(self, request: RESTRequest) -> RESTRequest:
        """
        Adds the server time and the signature to the request, required for authenticated interactions. It also adds
        the required parameter in the request header.

        :param request: the request to be configured for authenticated interaction
        """

        headers = {}
        if request.headers is not None:
            headers.update(request.headers)
        headers.update(self.authentication_headers(request=request))
        request.headers = headers

        return request

    async def ws_authenticate(self, request: WSRequest) -> WSRequest:
        """
        This method is intended to configure a websocket request to be authenticated. KuCoin does not use this
        functionality
        """
        return request  # pass-through

    def partner_header(self, timestamp: str):
        partner_payload = timestamp + CONSTANTS.HB_PARTNER_ID + self.api_key
        partner_signature = base64.b64encode(
            hmac.new(
                CONSTANTS.HB_PARTNER_KEY.encode("utf-8"),
                partner_payload.encode("utf-8"),
                hashlib.sha256).digest())
        third_party = {
            "KC-API-PARTNER": CONSTANTS.HB_PARTNER_ID,
            "KC-API-PARTNER-SIGN": str(partner_signature, "utf-8")
        }
        return third_party

    def authentication_headers(self, request: RESTRequest) -> Dict[str, Any]:
        timestamp = int(self.time_provider.time() * 1000)
        header = {
            "KC-API-KEY": self.api_key,
            "KC-API-TIMESTAMP": str(timestamp),
            "KC-API-KEY-VERSION": "2"
        }

        path_url = f"/api{request.url.split('/api')[-1]}"
        if request.params:
            sorted_params = self.keysort(request.params)
            query_string_components = urlencode(sorted_params, safe=',')
            path_url = f"{path_url}?{query_string_components}"

        if request.data is not None:
            body = request.data
        else:
            body = ""
        payload = str(timestamp) + request.method.value.upper() + path_url + body

        signature = base64.b64encode(
            hmac.new(
                self.secret_key.encode("utf-8"),
                payload.encode("utf-8"),
                hashlib.sha256).digest())
        passphrase = base64.b64encode(
            hmac.new(
                self.secret_key.encode('utf-8'),
                self.passphrase.encode('utf-8'),
                hashlib.sha256).digest())
        header["KC-API-SIGN"] = str(signature, "utf-8")
        header["KC-API-PASSPHRASE"] = str(passphrase, "utf-8")
        partner_headers = self.partner_header(str(timestamp))
        header.update(partner_headers)
        return header



import asyncio
from collections import defaultdict
from decimal import Decimal
from typing import TYPE_CHECKING, Any, AsyncIterable, Dict, List, Optional, Tuple

from bidict import bidict

from hummingbot.connector.constants import s_decimal_NaN
from hummingbot.connector.derivative.bitmart_perpetual import (
    bitmart_perpetual_constants as CONSTANTS,
    bitmart_perpetual_web_utils as web_utils,
)
from hummingbot.connector.derivative.bitmart_perpetual.bitmart_perpetual_api_order_book_data_source import (
    BitmartPerpetualAPIOrderBookDataSource,
)
from hummingbot.connector.derivative.bitmart_perpetual.bitmart_perpetual_auth import BitmartPerpetualAuth
from hummingbot.connector.derivative.bitmart_perpetual.bitmart_perpetual_user_stream_data_source import (
    BitmartPerpetualUserStreamDataSource,
)
from hummingbot.connector.derivative.position import Position
from hummingbot.connector.perpetual_derivative_py_base import PerpetualDerivativePyBase
from hummingbot.connector.trading_rule import TradingRule
from hummingbot.connector.utils import combine_to_hb_trading_pair
from hummingbot.core.api_throttler.data_types import RateLimit
from hummingbot.core.data_type.common import OrderType, PositionAction, PositionMode, PositionSide, PriceType, TradeType
from hummingbot.core.data_type.in_flight_order import InFlightOrder, OrderState, OrderUpdate, TradeUpdate
from hummingbot.core.data_type.order_book_tracker_data_source import OrderBookTrackerDataSource
from hummingbot.core.data_type.trade_fee import TokenAmount, TradeFeeBase
from hummingbot.core.data_type.user_stream_tracker_data_source import UserStreamTrackerDataSource
from hummingbot.core.utils.async_utils import safe_gather
from hummingbot.core.utils.estimate_fee import build_perpetual_trade_fee
from hummingbot.core.web_assistant.web_assistants_factory import WebAssistantsFactory

if TYPE_CHECKING:
    from hummingbot.client.config.config_helpers import ClientConfigAdapter

bpm_logger = None


class BitmartPerpetualDerivative(PerpetualDerivativePyBase):
    web_utils = web_utils
    SHORT_POLL_INTERVAL = 5.0
    UPDATE_ORDER_STATUS_MIN_INTERVAL = 10.0
    LONG_POLL_INTERVAL = 120.0

    def __init__(
            self,
            client_config_map: "ClientConfigAdapter",
            bitmart_perpetual_api_key: str = None,
            bitmart_perpetual_api_secret: str = None,
            bitmart_perpetual_memo: str = None,
            trading_pairs: Optional[List[str]] = None,
            trading_required: bool = True,
            domain: str = CONSTANTS.DOMAIN,
    ):
        self.bitmart_perpetual_api_key = bitmart_perpetual_api_key
        self.bitmart_perpetual_secret_key = bitmart_perpetual_api_secret
        self.bitmart_perpetual_memo = bitmart_perpetual_memo
        self._trading_required = trading_required
        self._trading_pairs = trading_pairs
        self._domain = domain
        self._position_mode = None
        self._last_trade_history_timestamp = None
        self._contract_sizes = {}
        super().__init__(client_config_map)

    @property
    def name(self) -> str:
        return CONSTANTS.EXCHANGE_NAME

    @property
    def authenticator(self) -> BitmartPerpetualAuth:
        return BitmartPerpetualAuth(api_key=self.bitmart_perpetual_api_key,
                                    api_secret=self.bitmart_perpetual_secret_key,
                                    memo=self.bitmart_perpetual_memo,
                                    time_provider=self._time_synchronizer)

    @property
    def rate_limits_rules(self) -> List[RateLimit]:
        return CONSTANTS.RATE_LIMITS

    @property
    def domain(self) -> str:
        return self._domain

    @property
    def client_order_id_max_length(self) -> int:
        return CONSTANTS.MAX_ORDER_ID_LEN

    @property
    def client_order_id_prefix(self) -> str:
        return CONSTANTS.BROKER_ID

    @property
    def trading_rules_request_path(self) -> str:
        return CONSTANTS.EXCHANGE_INFO_URL

    @property
    def trading_pairs_request_path(self) -> str:
        return CONSTANTS.EXCHANGE_INFO_URL

    @property
    def check_network_request_path(self) -> str:
        return CONSTANTS.SERVER_TIME_PATH_URL

    @property
    def trading_pairs(self):
        return self._trading_pairs

    @property
    def is_cancel_request_in_exchange_synchronous(self) -> bool:
        return True

    @property
    def is_trading_required(self) -> bool:
        return self._trading_required

    @property
    def funding_fee_poll_interval(self) -> int:
        return 600

    def supported_order_types(self) -> List[OrderType]:
        """
        :return a list of OrderType supported by this connector
        """
        return [OrderType.LIMIT, OrderType.MARKET, OrderType.LIMIT_MAKER]

    def supported_position_modes(self):
        """
        This method needs to be overridden to provide the accurate information depending on the exchange.
        """
        return [PositionMode.HEDGE]

    def get_buy_collateral_token(self, trading_pair: str) -> str:
        trading_rule: TradingRule = self._trading_rules[trading_pair]
        return trading_rule.buy_order_collateral_token

    def get_sell_collateral_token(self, trading_pair: str) -> str:
        trading_rule: TradingRule = self._trading_rules[trading_pair]
        return trading_rule.sell_order_collateral_token

    def _is_request_exception_related_to_time_synchronizer(self, request_exception: Exception):
        error_description = str(request_exception)
        is_time_synchronizer_related = ("40039" in error_description
                                        and "The timestamp is invalid" in error_description)
        return is_time_synchronizer_related

    def _is_order_not_found_during_status_update_error(self, status_update_exception: Exception) -> bool:
        return str(CONSTANTS.ORDER_NOT_EXIST_ERROR_CODE) in str(
            status_update_exception
        ) and CONSTANTS.ORDER_NOT_EXIST_MESSAGE in str(status_update_exception)

    def _is_order_not_found_during_cancelation_error(self, cancelation_exception: Exception) -> bool:
        return str(CONSTANTS.UNKNOWN_ORDER_ERROR_CODE) in str(
            cancelation_exception
        ) and CONSTANTS.UNKNOWN_ORDER_MESSAGE in str(cancelation_exception)

    def _create_web_assistants_factory(self) -> WebAssistantsFactory:
        return web_utils.build_api_factory(
            throttler=self._throttler,
            time_synchronizer=self._time_synchronizer,
            domain=self._domain,
            auth=self._auth)

    def _create_order_book_data_source(self) -> OrderBookTrackerDataSource:
        return BitmartPerpetualAPIOrderBookDataSource(
            trading_pairs=self._trading_pairs,
            connector=self,
            api_factory=self._web_assistants_factory,
            domain=self.domain,
        )

    def _create_user_stream_data_source(self) -> UserStreamTrackerDataSource:
        return BitmartPerpetualUserStreamDataSource(
            auth=self._auth,
            connector=self,
            api_factory=self._web_assistants_factory,
            domain=self.domain,
        )

    def _format_amount_to_size(self, trading_pair, amount: Decimal) -> Decimal:
        return int(amount / self._contract_sizes[trading_pair])

    def _format_size_to_amount(self, trading_pair, size: Decimal) -> Decimal:
        return Decimal(size) * self._contract_sizes[trading_pair]

    @property
    def side_mapping(self):
        return bidict(
            {
                (PositionAction.OPEN, TradeType.BUY): 1,  # buy_open_long
                (PositionAction.CLOSE, TradeType.BUY): 2,  # buy_close_short
                (PositionAction.CLOSE, TradeType.SELL): 3,  # sell_close_long
                (PositionAction.OPEN, TradeType.SELL): 4,  # sell_open_short
            }
        )

    @property
    def mode_mapping(self):
        return bidict(
            {
                OrderType.LIMIT: CONSTANTS.TIME_IN_FORCE_GTC,  # GTC
                OrderType.LIMIT_MAKER: CONSTANTS.TIME_IN_FORCE_MAKER_ONLY  # Maker only
            }
        )

    def _get_fee(self,
                 base_currency: str,
                 quote_currency: str,
                 order_type: OrderType,
                 order_side: TradeType,
                 position_action: PositionAction,
                 amount: Decimal,
                 price: Decimal = s_decimal_NaN,
                 is_maker: Optional[bool] = None) -> TradeFeeBase:
        is_maker = is_maker or False
        fee = build_perpetual_trade_fee(
            self.name,
            is_maker,
            base_currency=base_currency,
            quote_currency=quote_currency,
            order_type=order_type,
            order_side=order_side,
            position_action=position_action,
            amount=amount,
            price=price,
        )
        return fee

    async def _update_trading_fees(self):
        """
        Update fees information from the exchange
        """
        pass

    async def _place_cancel(self, order_id: str, tracked_order: InFlightOrder):
        symbol = await self.exchange_symbol_associated_to_pair(trading_pair=tracked_order.trading_pair)
        api_params = {
            "client_order_id": order_id,
            "symbol": symbol,
        }
        cancel_result = await self._api_post(
            path_url=CONSTANTS.CANCEL_ORDER_URL,
            data=api_params,
            is_auth_required=True)
        unknown_order_code = cancel_result.get("code") == CONSTANTS.UNKNOWN_ORDER_ERROR_CODE
        unknown_order_msg = cancel_result.get("msg", "") == CONSTANTS.UNKNOWN_ORDER_MESSAGE
        if unknown_order_msg and unknown_order_code:
            self.logger().debug(f"The order {order_id} does not exist on Bitmart Perpetual. "
                                f"No cancelation needed.")
            await self._order_tracker.process_order_not_found(order_id)
            raise IOError(f"{cancel_result.get('code')} - {cancel_result['msg']}")
        return cancel_result.get("code") == CONSTANTS.CODE_OK

    async def _place_order(
            self,
            order_id: str,
            trading_pair: str,
            amount: Decimal,
            trade_type: TradeType,
            order_type: OrderType,
            price: Decimal,
            position_action: PositionAction = PositionAction.NIL,
            **kwargs,
    ) -> Tuple[str, float]:
        price_str = f"{price:f}"
        symbol = await self.exchange_symbol_associated_to_pair(trading_pair=trading_pair)
        api_params = {
            "symbol": symbol,
            "client_order_id": order_id,
            "type": "market" if order_type == OrderType.MARKET else "limit",
            "side": self.side_mapping.get((position_action, trade_type)),
            "mode": self.mode_mapping.get(order_type),
            "size": self._format_amount_to_size(trading_pair, amount),
        }
        if order_type.is_limit_type():
            api_params["price"] = price_str
        order_result = await self._api_post(
            path_url=CONSTANTS.SUBMIT_ORDER_URL,
            data=api_params,
            is_auth_required=True)
        response_code = order_result.get("code")
        if response_code != 1000:
            raise IOError(f"Error submitting order {order_id}: {order_result['message']}")
        o_id = str(order_result["data"]["order_id"])
        transact_time = self._time_synchronizer.time()
        return o_id, transact_time

    async def _request_order_status(self, tracked_order: InFlightOrder) -> OrderUpdate:
        order_update = await self._api_get(
            path_url=CONSTANTS.ORDER_DETAILS,
            params={
                "symbol": await self.exchange_symbol_associated_to_pair(tracked_order.trading_pair),
                "order_id": tracked_order.exchange_order_id
            },
            is_auth_required=True)
        if order_update["code"] != 1000:
            if self._is_request_exception_related_to_time_synchronizer(request_exception=order_update):
                _order_update = OrderUpdate(
                    trading_pair=tracked_order.trading_pair,
                    update_timestamp=self.current_timestamp,
                    new_state=tracked_order.current_state,
                    client_order_id=tracked_order.client_order_id,
                )
                return _order_update
        order_update_data = order_update.get("data")
        if order_update is not None:
            deal_size = Decimal(order_update_data["deal_size"])
            size = Decimal(order_update_data["size"])
            state = order_update_data["state"]
            order_state = self.get_order_state(size, state, deal_size)
            _order_update: OrderUpdate = OrderUpdate(
                trading_pair=tracked_order.trading_pair,
                update_timestamp=order_update_data["update_time"] * 1e-3,
                new_state=order_state,
                client_order_id=order_update_data["client_order_id"],
                exchange_order_id=order_update_data["order_id"],
            )
            return _order_update

    async def _iter_user_event_queue(self) -> AsyncIterable[Dict[str, any]]:
        while True:
            try:
                yield await self._user_stream_tracker.user_stream.get()
            except asyncio.CancelledError:
                raise
            except Exception:
                self.logger().network(
                    "Unknown error. Retrying after 1 seconds.",
                    exc_info=True,
                    app_warning_msg="Could not fetch user events from Bitmart. Check API key and network connection.",
                )
                await self._sleep(1.0)

    async def _user_stream_event_listener(self):
        """
        Wait for new messages from _user_stream_tracker.user_stream queue and processes them according to their
        message channels. The respective UserStreamDataSource queues these messages.
        """
        async for event_message in self._iter_user_event_queue():
            try:
                await self._process_user_stream_event(event_message)
            except asyncio.CancelledError:
                raise
            except Exception as e:
                self.logger().error(f"Unexpected error in user stream listener loop: {e}", exc_info=True)
                await self._sleep(5.0)

    @staticmethod
    def get_order_state(size: Decimal, state: int, deal_size: Decimal) -> OrderState:
        """
        Determines the state of an order based on the provided parameters. Valid until Bitmart implements spot order
        states.

        Args:
            size (Decimal): The total size of the order.
            state (int): The state code of the order (e.g., 2 for active, 4 for closed).
            deal_size (Decimal): The size of the order that has been executed.

        Returns:
            OrderState: The determined order state (OPEN, PARTIALLY_FILLED, CANCELED, FILLED).

        Raises:
            UnknownOrderStateException: If the order state is not tracked or does not match any known conditions.
        """
        if state == 2 and deal_size == 0:
            return OrderState.OPEN
        elif state == 2 and (0 < deal_size < size):
            return OrderState.PARTIALLY_FILLED
        elif state == 4 and deal_size < size:
            return OrderState.CANCELED
        elif state == 4 and deal_size == size:
            return OrderState.FILLED
        else:
            raise UnknownOrderStateException(state, size, deal_size)

    async def _process_user_stream_event(self, event_message: Dict[str, Any]):
        event_data = event_message.get("data", {})
        event_group: str = event_message.get("group", "")
        if CONSTANTS.WS_ORDERS_CHANNEL in event_group and bool(event_data):
            order_message = event_data[0].get("order")
            client_order_id = order_message.get("client_order_id", None)
            tracked_order = self._order_tracker.all_fillable_orders.get(client_order_id)
            position_side = order_message.get("side")
            position_action = self.side_mapping.inv[position_side][0]
            if tracked_order is not None:
                trades_dict = order_message.get("last_trade")
                if trades_dict is not None:
                    trade_id: str = str(trades_dict["lastTradeID"])
                    fee_asset = trades_dict.get("feeCcy", tracked_order.quote_asset)
                    fee_amount = Decimal(trades_dict.get("fee", "0"))
                    flat_fees = [] if fee_amount == Decimal("0") else [TokenAmount(amount=fee_amount, token=fee_asset)]
                    fee = TradeFeeBase.new_perpetual_fee(
                        fee_schema=self.trade_fee_schema(),
                        position_action=position_action,
                        percent_token=fee_asset,
                        flat_fees=flat_fees,
                    )
                    fill_base_amount = Decimal(self._format_size_to_amount(tracked_order.trading_pair,
                                                                           trades_dict["fillQty"]))
                    trade_update: TradeUpdate = TradeUpdate(
                        trade_id=trade_id,
                        client_order_id=client_order_id,
                        exchange_order_id=str(order_message["order_id"]),
                        trading_pair=tracked_order.trading_pair,
                        fill_timestamp=order_message["update_time"] * 1e-3,
                        fill_price=Decimal(trades_dict["fillPrice"]),
                        fill_base_amount=fill_base_amount,
                        fill_quote_amount=Decimal(str(fill_base_amount)) * Decimal(trades_dict["fillPrice"]),
                        fee=fee,
                    )
                    self._order_tracker.process_trade_update(trade_update)

            tracked_order = self._order_tracker.all_updatable_orders.get(client_order_id)
            if tracked_order is not None:
                deal_size = Decimal(order_message["deal_size"])
                size = Decimal(order_message["size"])
                state = order_message["state"]
                order_state = self.get_order_state(size, state, deal_size)
                order_update: OrderUpdate = OrderUpdate(
                    trading_pair=tracked_order.trading_pair,
                    update_timestamp=order_message["update_time"] * 1e-3,
                    new_state=order_state,
                    client_order_id=client_order_id,
                    exchange_order_id=str(order_message["order_id"]),
                )

                self._order_tracker.process_order_update(order_update)

        elif CONSTANTS.WS_ACCOUNT_CHANNEL in event_group and bool(event_data):
            asset_name = event_data["currency"]
            self._account_balances[asset_name] = Decimal(event_data["available_balance"]) + Decimal(event_data["frozen_balance"])
            self._account_available_balances[asset_name] = Decimal(event_data["available_balance"])

        elif CONSTANTS.WS_POSITIONS_CHANNEL in event_group and bool(event_data):
            for asset in event_data:
                trading_pair = asset["symbol"]
                try:
                    hb_trading_pair = await self.trading_pair_associated_to_exchange_symbol(trading_pair)
                    if hb_trading_pair in self.trading_pairs:
                        position_side = PositionSide["LONG" if asset['position_type'] == 1 else "SHORT"]
                        position = self._perpetual_trading.get_position(hb_trading_pair, position_side)
                        if position is not None:
                            amount = Decimal(asset["hold_volume"])
                            if amount == Decimal("0"):
                                pos_key = self._perpetual_trading.position_key(hb_trading_pair, position_side)
                                self._perpetual_trading.remove_position(pos_key)
                            else:
                                price = self.get_price_by_type(hb_trading_pair, PriceType.MidPrice)
                                bep = Decimal(asset["hold_avg_price"])
                                sign = 1 if position_side == PositionSide.LONG else -1
                                unrealized_pnl = Decimal(str(sign)) * (price / bep - 1)
                                position.update_position(position_side=position_side,
                                                         unrealized_pnl=unrealized_pnl,
                                                         entry_price=bep,
                                                         amount=Decimal(
                                                             "-1") * amount if position_side == PositionSide.SHORT else amount)
                        else:
                            await self._update_positions()
                except KeyError:
                    continue

    async def _format_trading_rules(self, exchange_info_dict: Dict[str, Any]) -> List[TradingRule]:
        """
        Queries the necessary API endpoint and initialize the TradingRule object for each trading pair being traded.

        Parameters
        ----------
        exchange_info_dict:
            Trading rules dictionary response from the exchange
        """
        rules: list = exchange_info_dict.get("data", [])
        return_val: list = []
        for rule in rules["symbols"]:
            try:
                if web_utils.is_exchange_information_valid(rule):
                    trading_pair = await self.trading_pair_associated_to_exchange_symbol(symbol=rule["symbol"])
                    contract_size = Decimal(rule.get("contract_size"))
                    self._contract_sizes[trading_pair] = contract_size
                    min_order_size_in_contracts = Decimal(rule.get("min_volume"))
                    min_order_size = contract_size * min_order_size_in_contracts
                    step_size = Decimal(rule.get("vol_precision"))
                    tick_size = Decimal(rule.get("price_precision"))
                    last_price = Decimal(rule.get("last_price"))
                    min_notional = Decimal(min_order_size * last_price)
                    collateral_token = rule["quote_currency"]

                    return_val.append(
                        TradingRule(
                            trading_pair,
                            min_order_size=min_order_size,
                            min_price_increment=Decimal(tick_size),
                            min_base_amount_increment=Decimal(step_size),
                            min_notional_size=Decimal(min_notional),
                            buy_order_collateral_token=collateral_token,
                            sell_order_collateral_token=collateral_token,
                        )
                    )
            except Exception as e:
                self.logger().error(
                    f"Error parsing the trading pair rule {rule}. Error: {e}. Skipping...", exc_info=True
                )
        return return_val

    def _initialize_trading_pair_symbols_from_exchange_info(self, exchange_info: Dict[str, Any]):
        mapping = bidict()
        symbols_data = exchange_info.get("data", {})
        for symbol_data in filter(web_utils.is_exchange_information_valid, symbols_data.get("symbols", [])):
            exchange_symbol = symbol_data["base_currency"] + symbol_data["quote_currency"]
            base = symbol_data["base_currency"]
            quote = symbol_data["quote_currency"]
            trading_pair = combine_to_hb_trading_pair(base, quote)
            if trading_pair in mapping.inverse:
                self._resolve_trading_pair_symbols_duplicate(mapping, exchange_symbol, base, quote)
            else:
                mapping[exchange_symbol] = trading_pair
        self._set_trading_pair_symbol_map(mapping)

    async def _get_last_traded_price(self, trading_pair: str) -> float:
        exchange_symbol = await self.exchange_symbol_associated_to_pair(trading_pair=trading_pair)
        params = {"symbol": exchange_symbol}
        response = await self._api_get(
            path_url=CONSTANTS.EXCHANGE_INFO_URL,
            params=params)
        price = float(response["last_price"])
        return price

    async def get_last_traded_prices(self, trading_pairs: List[str] = None) -> Dict[str, float]:
        response = await self._api_get(path_url=CONSTANTS.EXCHANGE_INFO_URL)
        symbol_map = await self.trading_pair_symbol_map()
        last_traded_prices = {
            await self.trading_pair_associated_to_exchange_symbol(ticker["symbol"]): float(ticker["last_price"])
            for ticker in response["data"]["symbols"] if ticker["symbol"] in symbol_map.keys()
        }
        return last_traded_prices

    def _resolve_trading_pair_symbols_duplicate(self, mapping: bidict, new_exchange_symbol: str, base: str, quote: str):
        """Resolves name conflicts provoked by futures contracts.

        If the expected BASEQUOTE combination matches one of the exchange symbols, it is the one taken, otherwise,
        the trading pair is removed from the map and an error is logged.
        """
        expected_exchange_symbol = f"{base}{quote}"
        trading_pair = combine_to_hb_trading_pair(base, quote)
        current_exchange_symbol = mapping.inverse[trading_pair]
        if current_exchange_symbol == expected_exchange_symbol:
            pass
        elif new_exchange_symbol == expected_exchange_symbol:
            mapping.pop(current_exchange_symbol)
            mapping[new_exchange_symbol] = trading_pair
        else:
            self.logger().error(
                f"Could not resolve the exchange symbols {new_exchange_symbol} and {current_exchange_symbol}")
            mapping.pop(current_exchange_symbol)

    async def _status_polling_loop_fetch_updates(self):
        await safe_gather(
            self._update_trade_history(),
            self._update_order_status(),
            self._update_balances(),
            self._update_positions(),
        )

    async def _update_balances(self):
        """
        Calls the REST API to update total and available balances.
        """
        local_asset_names = set(self._account_balances.keys())
        remote_asset_names = set()

        account_info = await self._api_get(path_url=CONSTANTS.ASSETS_DETAIL,
                                           is_auth_required=True)
        assets = account_info.get("data", [])
        for asset in assets:
            asset_name = asset.get("currency")
            available_balance = Decimal(asset.get("available_balance"))
            wallet_balance = Decimal(asset.get("equity"))
            self._account_available_balances[asset_name] = available_balance
            self._account_balances[asset_name] = wallet_balance
            remote_asset_names.add(asset_name)

        asset_names_to_remove = local_asset_names.difference(remote_asset_names)
        for asset_name in asset_names_to_remove:
            del self._account_available_balances[asset_name]
            del self._account_balances[asset_name]

    async def _update_positions(self):
        positions = await self._api_get(path_url=CONSTANTS.POSITION_INFORMATION_URL,
                                        is_auth_required=True)
        for position in positions["data"]:
            trading_pair = position.get("symbol")
            try:
                hb_trading_pair = await self.trading_pair_associated_to_exchange_symbol(trading_pair)
            except KeyError:
                # Ignore results for which their symbols is not tracked by the connector
                continue
            position_type = position["position_type"]
            position_side = PositionSide["LONG" if position_type == 1 else "SHORT"]
            unrealized_pnl = Decimal(position.get("unrealized_value"))
            entry_price = Decimal(position.get("entry_price"))
            amount = Decimal(position.get("current_amount"))
            leverage = Decimal(position.get("leverage"))
            pos_key = self._perpetual_trading.position_key(hb_trading_pair, position_side)
            if amount != 0:
                _position = Position(
                    trading_pair=await self.trading_pair_associated_to_exchange_symbol(trading_pair),
                    position_side=position_side,
                    unrealized_pnl=unrealized_pnl,
                    entry_price=entry_price,
                    amount=Decimal("-1") * amount if position_side == PositionSide.SHORT else amount,
                    leverage=leverage
                )
                self._perpetual_trading.set_position(pos_key, _position)
            else:
                self._perpetual_trading.remove_position(pos_key)

    async def _all_trade_updates_for_order(self, order: InFlightOrder) -> List[TradeUpdate]:
        # since current connector standard reimplemented _update_order_status this method is never reached
        pass

    async def _update_trade_history(self):
        last_tick = int(self._last_poll_timestamp / self.UPDATE_ORDER_STATUS_MIN_INTERVAL)
        current_tick = int(self.current_timestamp / self.UPDATE_ORDER_STATUS_MIN_INTERVAL)
        if current_tick > last_tick and len(self._order_tracker.active_orders) > 0:
            trading_pairs_to_order_map: Dict[str, Dict[str, Any]] = defaultdict(lambda: {})
            for order in self._order_tracker.active_orders.values():
                trading_pairs_to_order_map[order.trading_pair][order.exchange_order_id] = order
            trading_pairs = list(trading_pairs_to_order_map.keys())
            tasks = [
                self._api_get(
                    path_url=CONSTANTS.ACCOUNT_TRADE_LIST_URL,
                    params={"symbol": await self.exchange_symbol_associated_to_pair(trading_pair=trading_pair)},
                    is_auth_required=True,
                )
                for trading_pair in trading_pairs
            ]
            self.logger().debug(f"Polling for order fills of {len(tasks)} trading_pairs.")
            results = await safe_gather(*tasks, return_exceptions=True)
            for trades, trading_pair in zip(results, trading_pairs):
                order_map = trading_pairs_to_order_map.get(trading_pair)
                if isinstance(trades, Exception):
                    self.logger().network(
                        f"Error fetching trades update for the order {trading_pair}: {trades}.",
                        app_warning_msg=f"Failed to fetch trade update for {trading_pair}."
                    )
                    continue
                for trade in trades["data"]:
                    order_id = trade.get("order_id")
                    if order_id is not None and order_id in order_map:
                        tracked_order: InFlightOrder = order_map.get(order_id)
                        position_side = PositionSide.LONG if trade["side"] == 1 else PositionSide.SHORT
                        position_action = (PositionAction.OPEN
                                           if (tracked_order.trade_type is TradeType.BUY and position_side == "LONG"
                                               or tracked_order.trade_type is TradeType.SELL and position_side == "SHORT")
                                           else PositionAction.CLOSE)
                        quote_asset = trading_pair.split("-")[1]
                        fee_amount = Decimal(trade["paid_fees"])
                        fee = TradeFeeBase.new_perpetual_fee(
                            fee_schema=self.trade_fee_schema(),
                            position_action=position_action,
                            percent_token=quote_asset,
                            flat_fees=[TokenAmount(amount=fee_amount, token=quote_asset)]
                        )
                        fill_base_amount = Decimal(self._format_size_to_amount(tracked_order.trading_pair,
                                                                               trade["vol"]))
                        trade_update: TradeUpdate = TradeUpdate(
                            trade_id=str(trade["trade_id"]),
                            client_order_id=tracked_order.client_order_id,
                            exchange_order_id=trade["order_id"],
                            trading_pair=tracked_order.trading_pair,
                            fill_timestamp=trade["create_time"] * 1e-3,
                            fill_price=Decimal(trade["price"]),
                            fill_base_amount=fill_base_amount,
                            fill_quote_amount=Decimal(str(fill_base_amount)) * Decimal(trade["price"]),
                            fee=fee,
                        )
                        self._order_tracker.process_trade_update(trade_update)

    async def _update_order_status(self):
        """
        Calls the REST API to get order/trade updates for each in-flight order.
        """
        last_tick = int(self._last_poll_timestamp / self.UPDATE_ORDER_STATUS_MIN_INTERVAL)
        current_tick = int(self.current_timestamp / self.UPDATE_ORDER_STATUS_MIN_INTERVAL)
        if current_tick > last_tick and len(self._order_tracker.active_orders) > 0:
            tracked_orders = list(self._order_tracker.active_orders.values())
            tasks = [
                self._api_get(
                    path_url=CONSTANTS.ORDER_DETAILS,
                    params={
                        "symbol": await self.exchange_symbol_associated_to_pair(trading_pair=order.trading_pair),
                        "order_id": order.exchange_order_id
                    },
                    is_auth_required=True,
                    return_err=True,
                )
                for order in tracked_orders
            ]
            self.logger().debug(f"Polling for order status updates of {len(tasks)} orders.")
            results: List[Dict[str, Any]] = await safe_gather(*tasks, return_exceptions=True)

            for order_update, tracked_order in zip(results, tracked_orders):
                client_order_id = tracked_order.client_order_id
                if client_order_id not in self._order_tracker.all_orders:
                    continue
                if isinstance(order_update, Exception) or order_update["code"] != 1000:
                    not_found_error = (order_update["code"] in (CONSTANTS.UNKNOWN_ORDER_ERROR_CODE,
                                                                CONSTANTS.UNKNOWN_ORDER_ERROR_CODE))
                    if not isinstance(order_update, Exception) and not_found_error:
                        await self._order_tracker.process_order_not_found(client_order_id)
                    else:
                        self.logger().network(
                            f"Error fetching status update for the order {client_order_id}: " f"{order_update}."
                        )
                    continue
                order_update_data = order_update["data"]
                size = Decimal(order_update_data["size"])
                deal_size = Decimal(order_update_data["deal_size"])
                state = order_update_data["state"]
                order_state = self.get_order_state(size, state, deal_size)
                new_order_update: OrderUpdate = OrderUpdate(
                    trading_pair=await self.trading_pair_associated_to_exchange_symbol(order_update_data['symbol']),
                    update_timestamp=order_update_data["update_time"] * 1e-3,
                    new_state=order_state,
                    client_order_id=order_update_data["client_order_id"],
                    exchange_order_id=order_update_data["order_id"],
                )

                self._order_tracker.process_order_update(new_order_update)

    async def _trading_pair_position_mode_set(self, mode: PositionMode, trading_pair: str) -> Tuple[bool, str]:
        # TODO: Currently there are no position mode settings in Bitmart
        return True, ""

    async def _set_trading_pair_leverage(self, trading_pair: str, leverage: int) -> Tuple[bool, str]:
        symbol = await self.exchange_symbol_associated_to_pair(trading_pair)
        leverage_str = str(leverage)
        # TODO: Check if there is something to handle cross/isolated
        payload = {"symbol": symbol, "leverage": leverage_str, "open_type": "cross"}
        set_leverage = await self._api_post(
            path_url=CONSTANTS.SET_LEVERAGE_URL,
            data=payload,
            is_auth_required=True,
        )
        success = False
        msg = ""
        if set_leverage["code"] == CONSTANTS.CODE_OK:
            success = set_leverage["data"]["leverage"] == leverage_str
        else:
            msg = 'Unable to set leverage'
        return success, msg

    async def _fetch_last_fee_payment(self, trading_pair: str) -> Tuple[int, Decimal, Decimal]:
        timestamp, funding_rate, payment = 0, Decimal("-1"), Decimal("-1")

        exchange_symbol = await self.exchange_symbol_associated_to_pair(trading_pair)
        payment_response = await self._api_get(
            path_url=CONSTANTS.GET_INCOME_HISTORY_URL,
            params={
                "symbol": exchange_symbol,
                "flow_type": 3,
            },
            is_auth_required=True,
        )
        payment_data = payment_response.get("data")
        funding_info_response = await self._api_get(
            path_url=CONSTANTS.FUNDING_INFO_URL,
            params={
                "symbol": exchange_symbol,
            },
        )
        if bool(payment_data):
            sorted_payment_response = sorted(payment_data, key=lambda a: a["time"], reverse=True)
            funding_payment = sorted_payment_response[0]
            payment = Decimal(funding_payment["amount"])
            timestamp = int(funding_payment["time"])
            funding_rate = Decimal(funding_info_response["data"]["rate_value"])
        return timestamp, funding_rate, payment


class UnknownOrderStateException(Exception):
    """Custom exception for unknown order states."""
    def __init__(self, state, size, deal_size):
        super().__init__(f"Order state {state} with size {size} and deal size {deal_size} not tracked. "
                         f"Please report this to a developer for review.")
